/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 1999
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/**	Column.m - class to store columns of data with further information.
 *
 *	Andreas Beyer, 1998
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <Foundation/NSException.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>

#include "Column.h"
#include "statist.h"
#include "sx_funcs.h"

@implementation Column
-init: (const double*) array size: (unsigned) _size label: (const char*) lb {
	[super init];
	size = _size;
	label  = calloc( strlen(lb)+1, sizeof(char) );
	column = calloc( size, sizeof( *column ) );
	if( !label || !column ) return NULL;
	strcpy( label, lb );
	sx_copy_d_array( column, array, size );

	return [self initProperties];
}

-initProperties {
// code adopted from statist
  	double var_coef, *xh, t;
  	double index, no_u, no_l;
  	int i, i_l, i_u, xcrit;

  /* kritische Werte fuer 95% Konfidenzinterval des Medians aus NEAVE: */
  int critical[50] = { 
    -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3,
    3, 4, 4, 4, 5, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11,
    11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 17, 17
  };

    	min = get_min(column, size);
    	max = get_max(column, size);


  	if (min > max) {
		[NSException raise: NSInvalidArgumentException
			format: @"-initProperties: min > max"];
  	}
    
  	mean = get_mean(column, size);
  	std_var = get_sdv(column, size);
  	stdv_mean = std_var/sqrt((double) size);
  	var_coef = (std_var/mean) * 100.;
  	t = get_t(0.975, (size-1));
  	conf = (t*std_var)/sqrt((double)size);
											
	  
  
  	xh = sx_calloc_double(size);
  	if( !xh ) [NSException raise: NSMallocException
		format: @"-initProperties: Not enough memory"];
	sx_copy_d_array( xh, column, size );
  
  	qsort(xh, size, sizeof(double), real_compar_up);
  	
  	if (size % 2 == 1) {
  	  i = (size-1)/2;
  	  median = xh[i];
  	}
  	else {
  	  i = (size/2) -1;
  	  median = xh[i];
  	  i++;
  	  median = (median + xh[i])/2.;
  	}
    
  	index = (double)size*0.25;
  	if (index==floor(index)) {
  	  i_l = (int)index-1;
  	  i_u = (int)index;
  	}
  	else {
  	  i_l = (int) floor(index-1.);
  	  i_u = (int) ceil(index-1.);
  	}
  
  	quart_l = 0.5 * (xh[i_l]+xh[i_u]);
  	quart_u = 0.5 * (xh[size-i_l-1]+xh[size-i_u-1]);
  	no_u = median + 1.58*(quart_u-quart_l)/sqrt((double)size);
  	no_l = median - 1.58*(quart_u-quart_l)/sqrt((double)size);
  
  	if (size<6) {
  	  med_conf_l = med_conf_u = 0.0;
  	}
  	else if (size<=50) {
  	  med_conf_l = xh[critical[size-1]];
  	  med_conf_u = xh[size-critical[size-1]-1];
  	}
  	else {
  	  xcrit = (int)floor(-(0.5*sqrt((double)size)) *1.96-0.5+0.5*(double)size);
  	  med_conf_l = xh[xcrit];
  	  med_conf_u = xh[size-xcrit-1];
  	}
	free( xh );
	return self;
}

-setLabel: (const char*) lb {
	if( label ) free( label );
	label = calloc( strlen(lb)+1, sizeof(char) );
	if( !label ) return NULL;
	strcpy( label, lb );

	return self;
}

-(unsigned) count {
	return size;
}

-(const char*) getLabel {
	return label;
}
// return data at position 'pos'
-(double) dataAt: (unsigned) pos {
	if( pos >= size )
		return 0;
	return column[pos];
}
// return the ORIGINAL array of data
-(double*) getArray {
	return column;
}

-copy {
	Column* res = [[[self class] alloc] 
			init: [self getArray] 
			size:  size
			label: label];	
	return res;
}
// All entries must be 1's or 0's
-(BOOL) isBinary {
	unsigned i;
	for( i = 0; i < size; i++ )
		if( column[i] != 0 && column[i] != 1 )
			return NO;
	return YES;
}
// Check weather all numbers are positive integers.
-(BOOL) hasPosIntegers {
	unsigned i;
	for( i = 0; i < size; i++ )
		if( column[i] < 0 || ((int)column[i]) != column[i] )
			return NO;
	return YES;
}
// returns YES if condition is true for all data in the column:
-(BOOL) allEntriesLessOrEqual: (double) a_value {
	unsigned i;
	for( i = 0; i < size; i++ )
		if( column[i] > a_value )
			return NO;
	return YES;
}
-(BOOL) allEntriesLargerOrEqual: (double) a_value {
	unsigned i;
	for( i = 0; i < size; i++ )
		if( column[i] < a_value )
			return NO;
	return YES;
}
-(BOOL) allEntriesLessThan: (double) a_value {
	unsigned i;
	for( i = 0; i < size; i++ )
		if( column[i] >= a_value )
			return NO;
	return YES;
}
-(BOOL) allEntriesLargerThan: (double) a_value {
	unsigned i;
	for( i = 0; i < size; i++ )
		if( column[i] <= a_value )
			return NO;
	return YES;
}

-(double) min {
	return min;
}
-(double) max {
	return max;
}

/* TRIPLE.value = mean, 
   TRIPLE.upper_b = mean + conf, 
   TRIPLE.lower_b = mean - conf
*/
-(TRIPLE) mean {
	TRIPLE res;
	res.value = mean;
	res.upper_b = mean + conf;
	res.lower_b = mean - conf;

	return res;
}
/* TRIPLE.value = median, 
   TRIPLE.upper_b = med_conf_u, 
   TRIPLE.lower_b = med_conf_l
*/
-(TRIPLE) median {
	TRIPLE res;
	res.value = median;
	res.upper_b = med_conf_u;
	res.lower_b = med_conf_l;

	return res;
}
-(double) upperQuartil {
	return quart_u;
}
-(double) lowerQuartil {
	return quart_l;
}
-(double) stdVariation {
	return std_var;
}
-(double) stdVarOfMean {
	return stdv_mean;
}

/* -(void) release {

fprintf(stderr, "Column %d: release is called.\tRetain: %d\n", 
(int) self, [self retainCount]);

	[super release];
} 
 */

-(void) dealloc {
#ifdef DEBUG
	fprintf(stderr, "col %s free\n", label);
#endif
	if( column) free( column );
	if( label ) free( label );
	[super dealloc];
}
@end
