/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 1999
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/**	DataDisplay.m - display data in a table.
 *
 *
 *	Andreas Beyer, 1998
 */

#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSException.h>
#include <Foundation/NSString.h>

#include <GToolKit/GToolKit.h>
#include <gdk/gdk.h>

#include <stdio.h>  // stderr, sdout, ...
#include <math.h>

#include "DataDisplay.h"
#include "DataSet.h"
#include "Column.h"
#include "OKDialog.h"
#include "Manager.h"
#include "Message.h"
#include "SXException.h"
#include "SXKernel.h"
#include "SXProgressBar.h"

/* Look at NSObject documentation, "methodForSelector:" 
 * Prototype for [ds getDataAtColumn: col row: row]
 */
typedef double (*SXGetDataIMP)(id, SEL, unsigned, unsigned);


@implementation DataDisplay 
+dataDisplayWithDataSet: (DataSet*) data_set {
	return [ (DataDisplay*) [[self alloc] autorelease]
		initWithDataSet: (DataSet*) data_set];
}

-initWithDataSet: (DataSet*) data_set {
	unsigned	col, row;
	// This contains the number of data in each column.
	unsigned	*data_n_rows;
	unsigned	n_row_titles;
	GtkSheet*	gtk_sheet;
	GtkSheetRange	range;
	NSArray*	row_titles;
	NSString*	row_title;
	SXGetDataIMP	getData;

	manager	= [[Manager getManager] retain];
	ds 		= [data_set retain];
	title 	= [[ds getFileName] lastPathComponent];
	[title retain]; // This shouldnt be necessary!
	n_cols = [ds countColumns];
	data_n_rows = malloc( n_cols*sizeof(*data_n_rows) );
	if(!data_n_rows)
		[NSException raise: NSMallocException
			format: @"-init: Can't create array for numbers of rows."];
	selected_cols = [[NSMutableArray alloc] initWithCapacity: n_cols];
	if(!selected_cols) {
		free( data_n_rows );
		[NSException raise: NSMallocException
			format: @"-init: Can't create array for selected columns."];
	}
	// get max. count of rows:
	n_rows = total_n_elements = 0;
	for ( col = 0; col < n_cols; col++ ) {
		data_n_rows[col] = [ds countRows: col];
		if ( data_n_rows[col] > n_rows ) n_rows = data_n_rows[col];
		total_n_elements += data_n_rows[col];
	}
	[super initBrowserWithRows: n_rows columns: n_cols title: title];
	gtk_sheet = GTK_SHEET ([self gtk]);

	// get row titles:
	row_titles = [ds getRowLabels];
	n_row_titles = [row_titles count];
	for( row = 0; row < n_row_titles; row++) {
		row_title = (NSString *) [row_titles objectAtIndex: row];
		[self setRow: row title:row_title];
		[self rowButtonAdd: row label:row_title];
	}
	for( ; row < n_rows; row++ ) {
		[self setRow: row title: [NSString stringWithFormat: @"%d", row]];
	}
	// set the whole sheet un-editable:
	range.row0=0;
 	range.rowi=n_rows;
 	range.col0=0;
 	range.coli=n_cols;
	[self rowsSetSensitivity: FALSE];
	[self setSelectionMode: GTK_SELECTION_SINGLE];

	
/*********** TABLE ******************/
	// fill the table:
	loading = YES; // Set this before creating the progress bar!
	getData = (SXGetDataIMP)
		[ds methodForSelector:@selector(getDataAtColumn:row:)];

	if( total_n_elements > 1000 )
		[SXProgressBar progressBarWithDelegate:self title: @"Drawing ..."];
	current_element = 0;
	// Freeze the visual update. (More efficient.)
	[self freeze];
	while (gtk_events_pending()) gtk_main_iteration();
    	for( col = 0; col < n_cols; col++ ) {
	   // set title of column
	   const char* label = [ds getLabelOfColumn: col];
	   NSString *s_label = [NSString stringWithCString: label];
	   double value;
	   gchar value_as_string[64];
	   [self setColumn: col title: s_label];
	   // the column titles have to be added separately:
	   [self columnButtonAdd:col label:s_label];

	   // INSERT VALUES
	   for( row = 0; row < data_n_rows[col]; row++) {
	   	current_element++;
		value = getData(ds, @selector(getDataAtColumn:row:), col, row);
		sprintf(value_as_string, "%12.4G", value);
		// this is *much* faster than obj-c
		gtk_sheet_set_cell_text	(gtk_sheet, row, col, value_as_string);
		if( !(current_element%100) ) {
		    while (gtk_events_pending()) gtk_main_iteration();
		}
	    }
	    while (gtk_events_pending()) gtk_main_iteration();
	    if( stop_loading ) {
			// cleanup
			[self release];
			[NSException raise: SXCancelOperationException
			format: @"-init: Loading canceled by user."];
			return NULL;
	    }
	}
	loading = NO;
	// Visual update ('unfreeze')
	[self thaw];
// set signals:
	[self connectSignal:@"select_column" withAction:
		[GTKAction actionWithTarget:self selector:@selector(doSelectColumn:)]];
	[self connectSignal:@"activate" withAction:
		[GTKAction actionWithTarget:self selector:@selector(doActivate)]];
	[self connectSignal:@"button_press_event" withTarget: self 
		sel:@selector(buttonPressEvent:data:) ];
	[self setColors];
	free( data_n_rows );
	return self;
}
// set colors for spreadsheet
-(void)setColors {
//	GtkWidget* e = [self getEntry];
	GtkWidget* e = [self gtk];
	#ifdef CELLS_BG_WHITE
	   bg_normal = malloc(sizeof( GdkColor ));
	   gdk_color_parse("white", bg_normal);
	   gdk_color_alloc(gtk_widget_get_colormap(e), bg_normal);
	#else
	   bg_normal   = &e->style->bg[GTK_STATE_NORMAL];
	#endif // CELLS_BG_WHITE
	bg_selected = &e->style->bg[GTK_STATE_SELECTED];
	fg_normal   = &e->style->fg[GTK_STATE_NORMAL];
	fg_selected = &e->style->fg[GTK_STATE_SELECTED];
	if(! bg_normal
	|| ! bg_selected
	|| ! fg_normal
	|| ! fg_selected)
		[NSException raise: NSMallocException
			format: @"-setColors: Color allocation failed."];
}
-(double) state {
	if( !loading ) return -1; // stop the progress bar
	return (double)current_element/(double)total_n_elements;
}
-(void) stopOperation {	// Called when the user selects "Cancel"
	stop_loading = YES;
}
-(NSString*) getTitle {
	// only give copy away:
	return [NSString stringWithString: title];
}

-(DataSet*)getDataSet {
	return ds;
}

-(GTKWidget *) getTable {
	return self;
}

// handle popup event
-(void)buttonPressEvent: sender data: data {
	gint x, y, row, col;
	GdkEventButton * event = (GdkEventButton*) data;
	GtkWidget	* sheet_w = GTK_WIDGET([self gtk]);
	GdkModifierType mods;

	gdk_window_get_pointer (sheet_w->window, NULL, NULL, &mods);
	if(mods&GDK_BUTTON3_MASK){ 
		x = event->x;
		y = event->y;
		[self getPixelInfo: x y: y row: &row column: &col];
		if( col >= n_cols || col < 0 ) return;
		[self raisePopupMenuForColumn: col event: event];
	}
}
-(void)unselectCurrentColumn {
	[self unselectColumnWithIndex: popup_target];
}
-(void)selectCurrentColumn {
	[self selectColumnWithIndex: popup_target];
}
-(void)histogramForCurrentColumn {
	Column * column = [ds getColumn: popup_target];
	SXKernel * kernel = [[Manager getManager] getKernel];
	[kernel histogramForCol: column
		  n_classes: 0 // if zero -> auto determination of classes
		  	  max: 0
			  min: 0 ];
}
-(void)outlierForCurrentColumn {
	Column * column = [ds getColumn: popup_target];
	SXKernel * kernel = [[Manager getManager] getKernel];
	[kernel outlierOfCol: column];
}
-(void)kolmogorovSmirnovForCurrentColumn {
	Column * column = [ds getColumn: popup_target];
	SXKernel * kernel = [[Manager getManager] getKernel];
	[kernel ksTestForCol: column];
}

-(void)raisePopupMenuForColumn: (unsigned) col event: (GdkEventButton*) event {
	GTKMenu	* menu;
	GTKMenuItem	* mItem;
	Column	* column;
	if( col >= n_cols ) return;
	popup_target = col;
	column = [ds getColumn: col];
	menu = [GTKMenu menu];
	if( [selected_cols containsObject: column] ) {
		// allow deselection
		mItem = [GTKMenuItem menuItemWithLabel: @"Unselect"];
		[mItem connectSignal:@"activate" withTarget: self 
			sel:@selector(unselectCurrentColumn) ];
		[menu append: mItem];
	} else {
		// allow selection
		mItem = [GTKMenuItem menuItemWithLabel: @"Select"];
		[mItem connectSignal:@"activate" withTarget: self 
			sel:@selector(selectCurrentColumn) ];
		[menu append: mItem];
	}
	// insert separator
	mItem = [GTKMenuItem menuItem];
	[menu append: mItem];
	mItem = [GTKMenuItem menuItemWithLabel: @"Histogram"];
	[mItem connectSignal:@"activate" withTarget: self 
		sel:@selector(histogramForCurrentColumn) ];
	[menu append: mItem];
	mItem = [GTKMenuItem menuItemWithLabel: @"Outlier"];
	[mItem connectSignal:@"activate" withTarget: self 
		sel:@selector(outlierForCurrentColumn) ];
	[menu append: mItem];
	mItem = [GTKMenuItem menuItemWithLabel: @"KS-Test"];
	[mItem connectSignal:@"activate" withTarget: self 
		sel:@selector(kolmogorovSmirnovForCurrentColumn) ];
	[menu append: mItem];

	[menu popup:NULL parentMenuItem:NULL func:NULL 
		data:NULL button:event->button activateTime:event->time];

}

-(void) doSelectColumn: (id) sender {
	Column* currentColumn;
	gint selected_col, selected_row;
	[self getActiveCell: &selected_row column: &selected_col];
	if( selected_col < 0 || selected_col > n_cols ) 
	// This should never be reached!
			return;
	currentColumn = [ds getColumn: selected_col];
	if( [selected_cols containsObject:  currentColumn] )
		[self unselectColumnWithIndex:selected_col];
	else
		[self selectColumnWithIndex:selected_col];
}
-(void) doActivate {
 	Column* currentColumn;
	gint selected_col, selected_row;
	[self getActiveCell: &selected_row column: &selected_col];
	currentColumn = [ds getColumn: selected_col];
//	if([selected_cols containsObject:  currentColumn]);
		[self unselectColumnWithIndex:selected_col]; 
}
// This method marks the selected column as selected
-(void)selectColumnWithIndex: (unsigned) index {
	Column* currentColumn;
	GtkSheetRange range;

	if( index < 0 || index > n_cols ) return;
	currentColumn = [ds getColumn: index];
	if(![selected_cols containsObject:  currentColumn])
		[selected_cols addObject: currentColumn];
	range.row0=0;
	range.rowi=n_rows;
	range.col0=index;
	range.coli=index;
	[self rangeSetBackground: &range color: bg_selected];
	[self rangeSetForeground: &range color: fg_selected];

}
-(void)unselectColumnWithIndex: (unsigned) index {
	Column* currentColumn;
	GtkSheetRange range;

	if( index < 0 || index > n_cols ) return;
	currentColumn = [ds getColumn: index];
	if([selected_cols containsObject:  currentColumn])
		[selected_cols removeObject:  currentColumn];
	range.row0=0;
	range.rowi=n_rows;
	range.col0=index;
	range.coli=index;
	[self rangeSetBackground: &range color: bg_normal];
	[self rangeSetForeground: &range color: fg_normal];
}
-(void)selectAllCols {
	unsigned i;
	for(i = 0; i < n_cols; i++) {
		if( ! [selected_cols containsObject: [ds getColumn: i]] )
			[self selectColumnWithIndex:i];
	}
}
-(void)unselectAllCols {
	unsigned i;
	for(i = 0; i < n_cols; i++) {
		if( [selected_cols containsObject: [ds getColumn: i]] )
			[self unselectColumnWithIndex:i];
	}
}

-(NSArray*)getSelectedCols {
	return selected_cols;
}


-(void)displayStandardData {
	unsigned	i, n;
	Column	*col;
	const char	*label;
	TRIPLE 	mean, median;
	OKDialog	*dialog;
	GTKLabel	*text;
	GTKHSeparator *h_sep;
//	GTKVSeparator *v_sep;
	GTKTable	*results;

	if( !ds ) return;
	n = [selected_cols count];
	if( !n ) {
		[Message messageWithContents: @"No columns selected!"
			title: @"Error"];
		return;
	}
    	results = [GTKTable tableWithRows: 7 columns: n+2 homogeneous: NO ];
	text = [GTKLabel labelWithStr: 
		[NSString stringWithFormat: @"Column"]];
	[text setAlignment: -1 yalign: 0];
	[results attach: text
		leftAttach: 0
		rightAttach: 1 
		topAttach: 0
		bottomAttach: 1
		xoptions: GTK_EXPAND | GTK_FILL
		yoptions: GTK_EXPAND
		xpadding: 1
		ypadding: 0
	];
	/* horizontal line under the columns' labels */
	for( i = 0; i < n+2; i++ ) {
		h_sep = [GTKHSeparator hSeparator];
		[results attach: h_sep
			leftAttach:   i
			rightAttach:  i+1 
			topAttach:    1
			bottomAttach: 2
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 0
			ypadding: 2
		];
	}
	text = [GTKLabel labelWithStr: 
		[NSString stringWithFormat: @"Nmb. of Elem."]];
	[text setAlignment: -1 yalign: 0];
	[results attach: text
		leftAttach: 0
		rightAttach: 1 
		topAttach: 2
		bottomAttach: 3
		xoptions: GTK_EXPAND | GTK_FILL
		yoptions: GTK_EXPAND
		xpadding: 1
		ypadding: 0
	];
	text = [GTKLabel labelWithStr: 
		[NSString stringWithFormat: @"Mean"]];
	[text setAlignment: -1 yalign: 0];
	[results attach: text
		leftAttach: 0
		rightAttach: 1 
		topAttach: 3
		bottomAttach: 4
		xoptions: GTK_EXPAND | GTK_FILL
		yoptions: GTK_EXPAND
		xpadding: 1
		ypadding: 0
	];
	text = [GTKLabel labelWithStr: 
		[NSString stringWithFormat: @"Median"]];
	[text setAlignment: -1 yalign: 0];
	[results attach: text
		leftAttach: 0
		rightAttach: 1 
		topAttach:    5
		bottomAttach: 6
		xoptions: GTK_EXPAND | GTK_FILL
		yoptions: GTK_EXPAND
		xpadding: 1
		ypadding: 0
	];
 	/* I would like to insert a vertical line after the rows' labels.
	   But that looks quite uggly. */
/*	for( i = 0; i < 5; i++ ) {
		v_sep = [GTKVSeparator vSeparator];
		[results attach:   v_sep
			leftAttach:  1
			rightAttach: 2 
			topAttach:	 i
			bottomAttach: i+1
			xoptions: GTK_EXPAND
			yoptions: GTK_EXPAND
			xpadding: 1
			ypadding: 1
		];
	}
 */	

	for( i = 0; i < n; i++ ) {
		col    = [selected_cols objectAtIndex: i];
		label  = [col getLabel];
		n_rows = [col count];
		median = [col median];
		mean   = [col mean];
		text = [GTKLabel labelWithStr: // name of the column
			[NSString stringWithFormat: @" %s ", label]];
		[results attach: text
			leftAttach:  i+1
			rightAttach: i+2 
			topAttach: 0
			bottomAttach: 1
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 2
			ypadding: 1
		];
		text = [GTKLabel labelWithStr: 
			[NSString stringWithFormat: @"%d", n_rows]];
		[results attach: text
			leftAttach:   i+1
			rightAttach:  i+2 
			topAttach:    2
			bottomAttach: 3
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 2
			ypadding: 1
		];
		text = [GTKLabel labelWithStr: 
			[NSString stringWithFormat: @"%.3f", mean.value]];
		[results attach: text
			leftAttach:   i+1
			rightAttach:  i+2 
			topAttach:    3
			bottomAttach: 4
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 2
			ypadding: 1
		];
		text = [GTKLabel labelWithStr:
			[NSString stringWithFormat: @"[%.2G; %.2G]", 
				mean.lower_b, mean.upper_b]];
		[results attach: text
			leftAttach:   i+1
			rightAttach:  i+2 
			topAttach:    4
			bottomAttach: 5
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 2
			ypadding: 1
		];
		text = [GTKLabel labelWithStr: 
			[NSString stringWithFormat: @"%.3f", median.value]];
		[results attach: text
			leftAttach:   i+1
			rightAttach:  i+2 
			topAttach:    5
			bottomAttach: 6
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 2
			ypadding: 1
		];
		text = [GTKLabel labelWithStr: 
			[NSString stringWithFormat: @"[%.2G; %.2G]", 
				median.lower_b, median.upper_b]];
		[results attach: text
			leftAttach:   i+1
			rightAttach:  i+2 
			topAttach:    6
			bottomAttach: 7
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 2
			ypadding: 1
		];

	}
	dialog = [OKDialog OKDialogWithContents: results title: [self getTitle]];
}

-(void) dealloc {
	[ds release]; // we said retain at the begining
	[selected_cols removeAllObjects];
	[selected_cols release];

	[manager release]; // we said retain at the begining

	[super dealloc];
#ifdef DEBUG
fprintf(stderr, "DD deallocated\n");
#endif
}
@end
