/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 1999
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/**	DataSet.m - store and evaluate data.
 *
 *
 *	Andreas Beyer, 1998
 */
#include <string.h>
#include <stdlib.h> // atof()
#include <errno.h>
#include <ctype.h>

#include <Foundation/NSException.h>
#include <Foundation/NSString.h>
#include <Foundation/NSScanner.h>

#include "DataSet.h"
#include "Column.h"
#include "Message.h"
#include "GraphView.h"
#include "SXException.h"
#include "SXFileHandle.h"
#include "SXSettings.h"

#include "sx_funcs.h"
#include "statist.h"	// statist header files

/* Look at NSObject documentation, "methodForSelector:"
 * Prototype for parseLine: (NSString*) line no: (unsigned) n_of_line
 */
typedef int (*SXParseLineIMP)(id, SEL, id, unsigned);

enum parse_line {
	pl_success,
	pl_skip_line,
	pl_stop_reading
};

char* get_label(double* x) {
// overwrite 'get_label()' from statist
	return [DataSet get_label: x];
}

// Add all created Columns to this list and remove them!
static NSMutableArray* all_objects;	// all Columns

@implementation DataSet
+(void) initialize {
	all_objects = [NSMutableArray new];
	if( ! all_objects )
		[NSException raise: NSMallocException
			format: @"Can't create list for columns."];
}

// Hack for statist
+(const char*) get_label: (double*) an_array {
	NSEnumerator *e = [all_objects objectEnumerator];
	Column* c;
	while( (c = [e nextObject]) ) {
		if( an_array == [c getArray] ) return [c getLabel];
	}
	return "(no label)";
}

-init {
	[super init];
	manager = [[Manager getManager] retain];
	return self;
}
-initWithFile: (NSString*) file_name {
	if(! [self init] ) return NULL;
	if( !file_name ) return self; // loadFile must be called later
	if( [self loadFile: file_name] ) return self;
	[self release]; // init didn't work
	return NULL;
}
-(NSArray*)componentsFromString:(NSString*)line separator: (NSString*) col_sep {
/** Create *new* array containing the compontents of line. */
	NSArray* t_array1;
	NSMutableArray* t_array2;
	unsigned n = 0;
	/* A line in a data file may contain several blanks after another 
	 * before the next actual number starts. Only the numbers are assigned
	 * to columns and blanks (if more than one) have to be ignored.
	 * Two column separators after another result in empty
	 * components and have to be ignored.
	 *
	 * This function creates a new NSArray with removed empty entries.
	 */
	if( !line || ![line length] ) return NULL;
	if( !col_sep ) {
		if( colSep ) col_sep = colSep;
		else	return NULL;
	}
	t_array1 = [line componentsSeparatedByString:col_sep];
	t_array2 = [[NSMutableArray alloc] initWithArray:t_array1];
	if( !t_array2 ) [NSException raise: NSMallocException
		format: @"-componentsFromString: Can't create NSArray."];
	while( n < [t_array2 count] ) { // components has at least one element
	    if( ![(NSString*)[t_array2 objectAtIndex: n] length] ) {
		 [t_array2 removeObjectAtIndex: n];
	    } else {
    		 n++;
    	    }
	}
	return t_array2;
}

-(int) parseLine_save: (NSString*) line no: (unsigned) n_of_line {
/** This method uses more Objective-C for parsing the line, which is
    safer than ANSI-C and more convenient for internationalization.
  */
	NSRange	comment;
	NSArray *	components; // components of the line
	NSScanner*	scanner;
	NSString*	msg;
	NSString*	component;
	NSString*	missingValue 
				= [NSString stringWithFormat: @"%c", SX_MISSING_VALUE];
	double	value;
	unsigned 	col, 
			skip_cols_cnt = 0; // position in 'skip_cols'

	if( !d_cols || !fh ) 		
		[NSException raise: NSGenericException
			format: @"-parseLine: Illegal state when calling this method."];

	// cut off comment
	comment = [line rangeOfString: @SX_COMMENT_INDICATOR];
	if( comment.length )
		line = [line substringToIndex: comment.location];
	if( [line length] == 0 ) {
		return pl_skip_line;
	}
	
	// separate columns
	/* The line may contain several blanks after another before
	 * the next actual number starts. Only the numbers are assigned
	 * to columns and blanks (if more than one) have to be ignored.
	 * Two column separators after another result in empty
	 * components and have to be ignored.
	 */
	components = [self componentsFromString:line separator: colSep];
	[components autorelease];
	if( [components count] < n_cols ) {
		msg = [NSString stringWithFormat:
			@"Too few entries in line %u!", [fh numberOfNextLine]-1];
		if( GTKRunMessageBox (@"StatistX: Warning", msg,
			@"Skip line", @"Cancel reading", nil) ) {
			return pl_stop_reading;
		} else { // skip line
			return pl_skip_line;
		}
	}
	if( n_of_line ) { // not in first line
		if( ! (n_of_line%SX_BUF_LEN) ) { // end of the arrays reached
			//Allocate longer arrays and copy data.
			sx_extend_array_length( d_cols, n_cols, n_of_line );
		}
	}
	// write data into temporary double-field
	for( col = 0; col < n_cols; col++ ) {
		if( n_skip_cols
		&&  skip_cols[skip_cols_cnt] == col ) {
			if( n_skip_cols > skip_cols_cnt ) skip_cols_cnt++;
			continue;
		}
		// interpret data
		component = (NSString*) [components objectAtIndex: col];
		if( row_labels && col == col_with_row_labels ) {
			[row_labels addObject: component];
			continue;
		}
		scanner = [NSScanner scannerWithString:component];
		if( [scanner scanDouble: &value] ) {
		/* This is the correct code:
			d_cols[col][n_of_line] = value;
		* But -scanDouble: is buggy implemented in libFoundation:
		* It parses exponents with more than one digit incorrectly.
		* Use the above code, once the bug is fixed.
		* Meanwhile use the following lines:
		*/
			value = atof([component cString]);
			if(errno == ERANGE) {
				msg = [NSString stringWithFormat:
					@"Illegal data in line %u.", [fh numberOfNextLine]-1];
				if( GTKRunMessageBox (@"StatistX: Warning", msg,
					@"Skip line", @"Cancel reading", nil) ) {
					return pl_stop_reading;
				} else { // skip line
					return pl_skip_line;
				}
			}
			d_cols[col][n_lines[col]] = value;
			n_lines[col]++;
		} else if( [component isEqualToString:missingValue] )
			continue;
		else {
			msg = [NSString stringWithFormat:
				@"Illegal data in line %u.", [fh numberOfNextLine]-1];
			if( GTKRunMessageBox (@"StatistX: Warning", msg,
				@"Skip line", @"Cancel reading", nil) ) {
				return pl_stop_reading;
			} else { // skip line
				return pl_skip_line;
			}
		}
	}
	return pl_success;
}

-(int) parseLine_fast: (NSString*) line no: (unsigned) n_of_line {
/** This method uses more ansi-C to parse the line, which is faster
    than parseLine_save:
    Use this method only if the column separators are blanks or tabs
    and if the data format follows English/American standards!
  */
	NSRange	comment;
	NSString*	msg;
	NSString*	row_label;
	const char*	c_line;
	const char* c_colSep = [colSep cString];
	char*		next = NULL;
	char*		tmp_s;
	double	value;
	unsigned 	col;
	// indicates current position in 'skip_cols' array:
	unsigned	skip_cols_cnt = 0;
	size_t		pos;

	if( !d_cols || !fh ) 		
		[NSException raise: NSGenericException
			format: @"-parseLine: Illegal state when calling this method."];

	// cut off comment
	comment = [line rangeOfString: @SX_COMMENT_INDICATOR];
	if( comment.length )
		line = [line substringToIndex: comment.location];
	if( [line length] == 0 ) {
		return pl_skip_line;
	}
	c_line = [line cString];
	if( n_of_line ) { // not in first line
		if( ! (n_of_line%SX_BUF_LEN) ) { // end of the arrays reached
			//Allocate longer arrays and copy data.
			sx_extend_array_length( d_cols, n_cols, n_of_line );
		}
	}
	// write data into temporary double-field
	for( col = 0; col < n_cols; col++ ) {
		/* The line may contain several blanks after another before
		 * the next actual number starts. Only the numbers are assigned
		 * to columns. Blanks (if more than one) have to be ignored.
		 */
		while(isspace( *c_line ) ) c_line++;
		if( !(*c_line) ) {
			msg = [NSString stringWithFormat:
				@"Too few data in line %u.", [fh numberOfNextLine]-1];
			if( SXRunMessageBox (@"StatistX: Warning", msg,
				@"Skip line", @"Cancel reading", nil) ) {
				return pl_stop_reading;
			} else { // skip line
				return pl_skip_line;
			}		
		}
		if( n_skip_cols
		&&  skip_cols[skip_cols_cnt] == col ) {
			c_line = strstr( c_line, c_colSep );
			if( n_skip_cols > skip_cols_cnt ) skip_cols_cnt++;
			continue;
		}
		if( row_labels && col == col_with_row_labels ) {
			pos = strcspn( c_line, c_colSep );
			tmp_s = calloc( sizeof(char), pos+1 );
			if( !tmp_s ) [NSException raise: NSMallocException
			format: @"-parseLine_fast: Not enough memory for row label."];
			tmp_s = strncpy(tmp_s, c_line, pos);
			row_label = [NSString stringWithCString: tmp_s];
			[row_labels addObject: row_label];
			free(tmp_s);
			c_line = c_line+pos;
			continue;
		}
		value = strtod(c_line, &next);
		if( next == c_line ) { // couldn't parse the string
			if( *c_line == SX_MISSING_VALUE ) {
			// SX_MISSING_VALUE must be one char
				c_line++; // skip this character
				continue;
			} else {
				msg = [NSString stringWithFormat:
					@"Illegal data format in line %u.", [fh numberOfNextLine]-1];
				if( SXRunMessageBox (@"StatistX: Warning", msg,
					@"Skip line", @"Cancel reading", nil) ) {
					return pl_stop_reading;
				} else { // skip line
					return pl_skip_line;
				}
			}
		
		}
		if(errno == ERANGE) {
			msg = [NSString stringWithFormat:
				@"Data out of range in line %u.", [fh numberOfNextLine]-1];
			if( SXRunMessageBox (@"StatistX: Warning", msg,
				@"Skip line", @"Cancel reading", nil) ) {
				return pl_stop_reading;
			} else { // skip line
				return pl_skip_line;
			}
		}
		c_line = next;
		d_cols[col][n_lines[col]] = value;
		n_lines[col]++;
	}
	return pl_success;
}

-(int) parseLine: (NSString*) line no: (int) no_of_line {
	return [self parseLine_save: line no: no_of_line];
}
-(BOOL)readDataFromStdin {
	fh = [SXFileHandle fileHandleWithCFilePointer: stdin];
	if( !fh ) return NO; // can't open file
	fileName = [[NSString stringWithFormat: @"stdin"] retain];
	return [self loadData];
}
-(BOOL) loadFile: (NSString*) path {
	fh = [SXFileHandle fileHandleWithPath: path];
	if( !fh ) return NO; // can't open file
	fileName = [path retain];
	return [self loadData];
}
-(BOOL)loadData {
	char*      t_label = NULL;
	const char* tmp_str = NULL;
	unsigned   i, col, n_labels = 0;
	unsigned   skip_cols_cnt = 0; // position in 'skip_cols'
	BOOL    no_data = YES;
	Column*    column = nil;
	NSArray *  components = NULL; // components of the line
	NSArray*   t_array = NULL;
	/* this array contains labels of columns, 
	   if labels are defined in data file     */
	NSArray*   labels_array = NULL; 
	NSRange    range, comment;
	NSString*  line = NULL, *labels_string = NULL, *label = NULL;
	NSString*  colSeps = [[SXSettings settings] columnSeparators];
	NSString*  t_colSep = NULL;

	SXParseLineIMP	parseLine;	// function pointer to method for parsing lines

	if(! fh) return NO;
	total_lines = [fh totalNumberOfLines];
	// find first line of data:
	line = nil;
	d_cols = NULL;
	do {	// skip all lines that start with a comment
		[line release];
		line = [fh getNextLine];
		if( !line ) {
			[Message messageWithContents: [NSString stringWithFormat: 
				@"File %@\nis empty or contains no data.", fileName]
				title: @"StatistX: Warning"
			];
			[line release];
			return NO;		
		}
		// cut off comment
		comment = [line rangeOfString: @SX_COMMENT_INDICATOR];
		if( comment.length ) {
			range = [line rangeOfString: @SX_LABEL_INDICATOR];
			if( range.length
			&& range.location == comment.location+comment.length ) {
			/* This comment line contains labels of the columns. We store
			   the line starting from the label indicator in labels_string
			   and have to extract the actual labels later, when we know,
			   what the column separator is.
			 */
				labels_string = [line substringFromIndex: range.location+range.length];
			}
			line  = [line substringToIndex: comment.location];
		}
	} while ( [line length] < 1 );

	/* Test which column separator fits best.
	   Policy: That separator is used, which produces most columns.
	 */
	range.length = 1;
	for( i = 0; i < [colSeps length]; i++ ) {
		range.location = i;
		t_colSep = [colSeps substringWithRange: range];
		t_array  = [self componentsFromString:line separator: t_colSep];
		[t_array autorelease];
		if( [t_array count] > n_cols ) {
			components = t_array;
			n_cols = [components count];
			colSep = t_colSep;
		}
	}
	/* Now
	   - colSep contains the chosen column separator
	   - n_cols contains the number of columns
	   - components contains the elements of the first row
	*/
	if( !components || !n_cols ) return NO;
	
	if( [colSep isEqualToString: @" "]
	||  [colSep isEqualToString: @"\t"] )
	    parseLine = (SXParseLineIMP)
		[self methodForSelector:@selector(parseLine_fast:no:)];
	else
	    parseLine = (SXParseLineIMP)
		[self methodForSelector:@selector(parseLine_save:no:)];
	
	/* Parse labels of the columns and find out which columns are to be skiped. */
	n_skip_cols = 0, skip_cols = NULL;
	if( labels_string ) { // break up the line into the labels
		skip_cols = malloc( n_cols * sizeof(*skip_cols) ) ;
		if( !skip_cols ) [NSException raise: NSMallocException
			format: @"-loadData: Not enough memory."];
		labels_array = [self componentsFromString:labels_string separator: colSep];
		[labels_array autorelease];
		n_labels = [labels_array count];
		if( !n_labels ) labels_array = NULL;
		labels_string = NULL; // Don't access this variable anymore!
		for( col = 0; col < n_labels; col++ ) {
			label = (NSString*) [labels_array objectAtIndex: col];
			if( [label characterAtIndex: 0] == SX_SKIP_COL_INDICATOR ) {
				skip_cols[n_skip_cols] = col;
				n_skip_cols++;
			}
			if( [label characterAtIndex: 0] == SX_ROW_LABEL_INDICATOR ) {
				row_labels = [NSMutableArray new];
				if( !row_labels )[NSException raise: NSMallocException
					format: @"-loadData: Not enough memory for row label array."];
				col_with_row_labels = col;
			}

		}
	}
	if( n_skip_cols == n_cols ) { // nothing to read
		[Message messageWithContents: 
				@"All columns skiped.\nNo data found."
			title: @"StatistX: Warning"];
		if( skip_cols ) free( skip_cols ), skip_cols = NULL;
		fh = NULL;
		if( row_labels ) [row_labels release];
		return NO;
	}
   NS_DURING
	// create temporary arrays:
	d_cols  = malloc( n_cols * sizeof( *d_cols  ) );
	n_lines = malloc( n_cols * sizeof( *n_lines ) );
	if( !d_cols || !n_lines )
		[NSException raise: NSMallocException
			format: @"-loadData: Not enough memory for column vector."];
	for( col = 0; col < n_cols; col++ ) {
	// sx_calloc_double checks whether malloc was successful for itself
		d_cols[col] = sx_calloc_double( SX_BUF_LEN );
		n_lines[col]= 0;
	}
	/************ Main loop - parsing data lines ***********/
	stop_operation = NO;
	if( total_lines > 20 )
		[SXProgressBar progressBarWithDelegate:self title: @"Reading data"];
	while (gtk_events_pending()) gtk_main_iteration();
	i = 0;
	do {
 		switch( (int)parseLine( self, @selector(parseLine:no:), line, i) ) 
		{
			case pl_success: break;
			case pl_skip_line: 
				[line release];
				continue;
			case pl_stop_reading:
			    if( skip_cols ) free( skip_cols );
			    if( row_labels ) [row_labels release];
			    if(d_cols) {
				for( col = 0; col < n_cols; col++ )
					free( d_cols[col] );	
				free( d_cols );
			    }
			    free( n_lines );
			    [line release];
			    fh = NULL;
			    return NO;
		}
		[line release];
		i++;
		current_line = [fh numberOfNextLine];
		while (gtk_events_pending()) gtk_main_iteration();
	} while( (line = [fh getNextLine]) 
	   &&    !stop_operation );
	/************ End parsing lines **************/
	no_data = ( i == 0 ); // i == number of read lines
	if( stop_operation ) {
		if( row_labels ) [row_labels release];
		if( skip_cols ) free( skip_cols ), skip_cols = NULL;
		for( col = 0; col < n_cols; col++ )
			free( d_cols[col] );	
		free( d_cols );
		free( n_lines );
		fh = NULL;
		return NO;
	}
	if( no_data ) {
	/* This piece of code is reached only, if all lines have
	   been skiped.
	 */
		[Message messageWithContents: 
				@"All lines skiped.\nNo data found."
			title: @"StatistX: Warning"];
		if( row_labels ) [row_labels release];
		if( skip_cols ) free( skip_cols ), skip_cols = NULL;
		for( col = 0; col < n_cols; col++ )
			free( d_cols[col] );	
		free( d_cols );
		free( n_lines );
		fh = NULL;
		return NO;
	}
	// fill actual Column objects with the data
	if( cols ) [cols release];
	cols = [[NSMutableArray alloc] initWithCapacity: n_cols];
	t_label = calloc( n_cols+1, sizeof( char ) );
	for( col = 0; col < n_cols; col++ ) {
		if( n_skip_cols
		&&  skip_cols[skip_cols_cnt] == col ) {
			if( n_skip_cols > skip_cols_cnt ) skip_cols_cnt++;
			continue;
		}
		if( row_labels && col == col_with_row_labels ) {
			continue;
		}
		if(  labels_array 
		 && [labels_array count] > col
		 && (labels_string = [labels_array objectAtIndex:col])
		 && [labels_string length]
		) {
		// label has been defined in data file
			[labels_string retain];
			column = [[Column alloc] init: d_cols[col] 
					size: n_lines[col] label:[labels_string cString]];
		} else {
		// create an own label for this column
			t_label[0] = '\0'; // reset string!
			t_label = sx_create_label( col, t_label );
			column = [[Column alloc] init: d_cols[col] 
					size: n_lines[col] label: t_label];		
		}
		if( ! column ) 
			[NSException raise: NSMallocException
				format: @"-loadFile: Not enough memory."];
		[cols addObject: column];	 // all columns in this DataSet
		[all_objects addObject: column]; // all columns ever created
		// TRICK: column gets a retain in cols.
		// Now, it will be automatically destroyed, if cols is destr.
		[column release];
	}

  NS_HANDLER
	// cleanup
	if( t_label ) free( t_label ), t_label = NULL;
	if( skip_cols ) free( skip_cols ), skip_cols = NULL;
	if( d_cols ) { 
		for( col = 0; col < n_cols; col++ )
			if( d_cols[col] ) free( d_cols[col] );	
		free( d_cols );
	}
	if( row_labels ) [row_labels release];
	if( n_lines ) free( n_lines ); n_lines = NULL;
	current_line = 0;
	fh = NULL;
	// re-raise the exception
	[localException raise];
  NS_ENDHANDLER

	// cleanup
	if( skip_cols ) free( skip_cols ), skip_cols = NULL;
	free( t_label );
	for( col = 0; col < n_cols; col++ )
		if( d_cols[col] ) free( d_cols[col] );	
	free( d_cols );
	fh = NULL;
	return YES;
}

// current state of file loading process
-(double) state {
	if( !fh || !n_cols || !total_lines ) return -1;
	return (double) current_line / (double) total_lines;
}
-(void)stopOperation {
	stop_operation = YES;
}


-(NSString*) getFileName {
	return [NSString stringWithString: fileName];
}

// not impl. -saveFile;
// not impl. -saveFileAs: (char*) file_name;
-(unsigned) countColumns {
	return [cols count];
}

// return number of rows in column number 'col'
-(unsigned) countRows: (unsigned) col {
	if( col >= n_cols )
		return 0;
	return [(Column*)[cols objectAtIndex: col] count];
}

-(Column*) getColumn: (unsigned) index {
	return [cols objectAtIndex: index];
}

-(unsigned) getIndexOfColumnWithLabel: (NSString*) label {
	unsigned index;
	const char *c_label = [label cString];
	const char* test_l;
	for( index = 0; index < n_cols; index++ ) {
		test_l = [(Column*) [cols objectAtIndex: index] getLabel];
		if( ! strcmp( c_label, test_l ) )
			return index;
	}
	[NSException raise: NSInvalidArgumentException 
		format: @"-getIndexOfColumnWithLabel: not found"];

	return index;
}

-(const char*) getLabelOfColumn: (unsigned) col {
	return [(Column*) [cols objectAtIndex: col] getLabel];
}

// return data at position col:row
-(double)   getDataAtColumn: (unsigned) col row: (unsigned) row {
	if( col < n_cols )
		return [(Column*) [cols objectAtIndex: col] dataAt: row];
	return 0;
}
-(NSArray*) getRowLabels {
	return row_labels;
}

-swapCols: (unsigned) col1 : (unsigned) col2 {
	Column* tmp;
	if( col1 >= n_cols || col2 >= n_cols )
		return NULL;
	tmp = [cols objectAtIndex: col1];
	[cols replaceObjectAtIndex: col1 withObject: [cols objectAtIndex: col2]];
	[cols replaceObjectAtIndex: col2 withObject: tmp];

	return self;
}

//==========================================================================
//>>>>>> statistical stuff:

//==================== Standard Menu =======================================

-(double) meanOfCol: (unsigned) col {
	Column* column;
	if( col >= n_cols ) 
		return 0;
	column = [cols objectAtIndex: col];
	// calling statist:
	return [column mean].value;
}

//============================= END STATISTCS ==============================
//==========================================================================

/* This is not the way it should be done:
-copy {
	unsigned i;
	DataSet * res = [[self class] alloc];
	res->manager  = manager;
	res->fileName = fileName;
	res->n_cols = n_cols;
	cols = [NSMutableArray arrayWithCapacity: n_cols];
	for( i = 0; i < n_cols; i++ ) {
		[res->cols addObject: [[cols objectAtIndex: i] copy]];
	}
	return res;
}
*/
-printStandardData: (unsigned) col {
	Column* c;
	if( col >= n_cols )
		return NULL;
		c = (Column*) [cols objectAtIndex: col];
fprintf(stderr, "calling statist\n");
	standard( [c  getArray], [c count], 0, 0, 0 );
	return self;
}

-printTable {
/** This may crash. For debugging only! */
	unsigned i,j, m;
	m = [(Column*) [cols objectAtIndex: 0] count];
	for (j = 0; j < m; j++){
		for (i = 0; i < n_cols; i++){
                        printf(" %6.2f", 
				[(Column*) [cols objectAtIndex: i] dataAt: j]);
                }
                putc('\n', stdout);
        }
	return self;
}

-(void) dealloc {
	unsigned i;
	[fileName release];
 	for( i = 0; i < n_cols; i++ ) {
		id obj = [cols objectAtIndex: i];
		[all_objects removeObject: obj];
	}
	[manager release];
	[cols release];
	if( row_labels ) [row_labels release];
	if( n_lines ) free( n_lines );
#ifdef DEBUG
fprintf(stderr, "DataSet deallocated.\n");
#endif // DEBUG
	[super dealloc];
	// statist:
	//finish_statist(); -> called in 'Manager'
	return;
}

@end
