/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 1999
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/**	GraphView.h - class to diplay (plot) graphs.
 *	Using gnuplot for this purpose.
 *
 *	Andreas Beyer, 1998
 */

#include <stdio.h>
#include <stdlib.h>

#include <Foundation/NSException.h>
#include <Foundation/NSFileManager.h>
#include <Foundation/NSString.h>
#include "GPPrintDialog.h"
#include "GraphView.h"
#include "Manager.h"
#include "LineDialog.h"
#include "SaveDialog.h"
#include "SXSettings.h"
#include "Message.h"

#include "statist.h"

// # gnuplot commands in histroy
#define N_CMD_HISTORY 5

@implementation GraphView
-init {
	Manager * manager;
	[super init];
	cmd_history = [[NSMutableArray alloc] initWithCapacity: N_CMD_HISTORY];
	manager = [Manager getManager];
	if( init_gnuplot() ) {
		[manager registerAction: @"/_Gnuplot/_Save Plot as EPS, color"
				target: self
				selector: @selector(saveCurrentAsEPS:)];
		[manager registerAction: @"/_Gnuplot/Save Plot as EPS, _monochrome"
				target: self
				selector: @selector(saveCurrentAsEPS_mono:)];
		[manager registerAction: @"/_Gnuplot/_Print Current Plot"
				target: self
				selector: @selector(printCurrent)];
		[manager registerAction: @"/_Gnuplot/_Replot"
				target: self
				selector: @selector(replot)];
		[manager registerAction: @"/_Gnuplot/Enter _Command"
				target: self
				selector: @selector(plotCommand:)];
	}
	return self;
}
// Save the current plot in a postscript file
-(void) saveCurrentAsEPS: sender {
	[self saveCurrentInFile:[SaveDialog runDialogWithInitialPath: nil]
		gpOptions: @"set terminal postscript eps enhanced color solid"];
}
-(void) saveCurrentAsEPS_mono: sender {
	[self saveCurrentInFile:[SaveDialog runDialogWithInitialPath: nil]
		gpOptions: @"set terminal postscript eps enhanced mono dashed"];
}

-(void) saveCurrentInFile: (NSString*) path 
		    gpOptions: (NSString*) gnuplot_options {

	if( !path || [path length] == 0 ) return;
/*	if( ! [[NSFileManager defaultManager] isWritableFileAtPath: path] ) {
		[Message messageWithContents: 
		[NSString stringWithFormat: @"Can't write to %@", path]
		 title: @"StatistX: Error"];
		return;
	}
*/
#ifdef DEBUG
	[(Manager*) GTKApp printOnDebug:
		[NSString stringWithFormat: @" saving text at %@ ...", path]];
#endif
	[self plotCommandWithCommand: [NSString stringWithFormat:
			@"%@ \nset output '%@' \nreplot \n set terminal x11", 
			gnuplot_options, path]
	];
}

-(void)printCurrent {
	BOOL	result;
	char	*line_style, *color_style;
	NSString* printerName;
	NSString* gp_options;
	GPPrintDialog* dialog =	[GPPrintDialog new];
	if( ! dialog ) 	[NSException raise: NSGenericException
		format: @"-printCurrent: Can't create GPPrintDialog"];
	result = [GTKApp runModalForWindow: dialog];
	if( !result ) {
		[dialog destroy];
		[dialog release];
		return;
	}
	printerName = [dialog getPrinterName];
	( [dialog getLineStyle]  == sx_solid ) 
		? (line_style = "solid") : (line_style = "dashed");
	( [dialog getColorStyle] == sx_color )
		? (color_style = "color") : (color_style = "mono");
	[dialog destroy];
	[dialog release];

	gp_options = [NSString stringWithFormat:
	  @"set terminal postscript eps enhanced %s %s", color_style, line_style];
	[self printCurrentOnPrinter: printerName gpOptions: gp_options];
}
/** Print the current plot to printer named 'printer' */
-(void) printCurrentOnPrinter: (NSString*) printer
		    gpOptions: (NSString*) gnuplot_options {
	NSFileManager * fileManager = [NSFileManager defaultManager];
	char buf[L_tmpnam], *tmp_name;
	NSString * print_cmd = [[SXSettings settings] valueForKey: @"PRINT_CMD"];
	NSString * cmd;
	int success;
	
	if( !printer ) return;
 	tmp_name = tmpnam( buf );
	if( !tmp_name ) [NSException raise: NSGenericException
		format: @"-printCurrentOnPrinter: Can't create temporary file name."];
	[self plotCommandWithCommand: [NSString stringWithFormat:
			@"%@ \nset output '%s' \nreplot \n set terminal x11", 
			gnuplot_options, tmp_name]
	];
	if( [print_cmd length] == 0 ) print_cmd = @"lpr";
	if( [printer length] == 0 ) {
	// take default printer, if [printer length] == 0
		cmd = [NSString stringWithFormat: 
			@"%@ %s >/dev/null 2>&1 ", print_cmd, tmp_name];
	} else {
		cmd = [NSString stringWithFormat: 
			@"%@ -P%@ %s >/dev/null 2>&1 ", print_cmd, printer, tmp_name];
	}
#ifdef DEBUG
fprintf(stderr, "print command: %s\n", [cmd cString] );
#endif
	success = system( [cmd cString] );
	if( success ) [Message messageWithContents:
		@"An error occured while printing."
		title: @"StatistX: Error"];
	// remove the temporary file:
	cmd = [NSString stringWithFormat: @"%s", tmp_name];
	[fileManager removeFileAtPath: cmd handler: NULL];
	
 	return;
}
// redraw
-(void)replot {
	[self plotCommandWithCommand: @"replot"];
}
-(void)plotCommand: sender {
	NSString *line;
	NSString *last_command;
	if (!init_gnuplot()) {
		[Message messageWithContents: @"Can't initialize Gnuplot."
		 title: @"StatistX: Error"];
		return;
	}
	if( [cmd_history count] > 0 ) {
		last_command = (NSString*) [cmd_history objectAtIndex: 0];
	} else {
		last_command = @"";
	}
	line = [LineDialog runDialogWithTitle:@"Gnuplot command"
			message: @"Enter a single Gnuplot command" 
			   text: last_command ];
	if( !line ) return;
	[self plotCommandWithCommand: line];
	[cmd_history insertObject: line atIndex: 0];
	if( [cmd_history count] > N_CMD_HISTORY )
		[cmd_history removeObjectAtIndex: N_CMD_HISTORY];
}

-(void) plotCommandWithCommand: (NSString*) gcommand {
	if (!init_gnuplot()
	||  ! gcommand
	||  [gcommand length] == 0) {
		return;
	}
	if ([[gcommand uppercaseString] isEqualToString: @"QUIT"] ) {
		pclose(pipef);
		gnupl_open = FALSE;
//		[Message messageWithContents: @"Sorry, you are not allowed \nto quit Gnuplot."
//		 title: @"StatistX: Error"];
		return;
	}
	fprintf(pipef, "%s\n", [gcommand cString]);
      fflush(pipef);
}
-(void) dealloc {
	[cmd_history release];
	[super dealloc];
}
@end
