/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 1999
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/**	Manager.m - catch (menu) events and handle them.
 *
 *
 *	Andreas Beyer, 1998
 */
#define IN_MANAGER_M TRUE // define this first

#include "Manager.h"

#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <GToolKit/GToolKit.h>

#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSFileManager.h>
#include <Foundation/NSException.h>
#include <Foundation/NSString.h>
#include <Foundation/NSCharacterSet.h>
#include <Foundation/NSArray.h>

#include "DataDisplay.h"
#include "FileList.h"
#include "GraphView.h"
#include "Message.h"
#include "OKDialog.h"
#include "TextPanel.h"
#include "SXHelp.h"
#include "SXKernel.h"
#include "SXAction.h"
#include "SXException.h"
#include "SXSettings.h"

#include "sx_defs.h"
#include "statist.h"	// statist header files

/* XPMs */
#include "ofolder.xpm"
#include "mean.xpm"
#include "printer.xpm"
#include "histo.xpm"
#include "linreg.xpm"
#include "normal.xpm"
#include "normal2.xpm"


static Manager *manager = 0;

static void get_sig(int sig) {
	[manager gotSignal: sig];
}

void finish_statist() {
// special finish-method for statist, don't call erasetempfiles()!
   if (log_set && logfile) {
     FCLOSE(logfile);
   }
   if (gnupl_open) {
     fprintf(pipef, "quit\n");
     pclose(pipef);
   }
   gnupl_open = FALSE;
   if (myexist(GPL_DAT)) {
     remove(GPL_DAT);
   }
}

@implementation Manager
// return the version of StatistX
+(NSString*)version {
	return @SX_VERSION;
}
// return version of statist
+(NSString*)statistVersion {
	return @_STATIST_VERSION_NUMBER;
}
+(Manager*)getManager {
	return manager;
}
- initWithArgc:(int *) argc argv:(char ***) argv {
	GTKVBox *		box;
	GTKHandleBox *	handle_box;
	GTKItemFactoryEntry menus[] = {
	// target = self
	   {{ "/_File", NULL, 0, 0, "<Branch>" }},
	   {{ "/File/_Open...",	"<control>O" },	@selector(menu_openFile:) },
	   {{ "/File/_Close", "<control>W" }, @selector(menu_closeFile:)},
	   {{ "/File/sep1", NULL, 0, 0, "<Separator>" }},
	   {{ "/File/_Quit", "<control>Q" }, @selector(terminate:)},
	   {{ "/File/sep2", NULL, 0, 0, "<Separator>" }},

	   {{ "/_Edit", NULL, 0, 0, "<Branch>" }},
	   {{ "/Edit/Select _All", "<control>A" },	@selector(menu_selectAll) },
	   {{ "/Edit/_Unselect All", NULL },	@selector(menu_unselectAll) },
	   {{ "/Edit/sep3", NULL, 0, 0, "<Separator>" }},

	   {{ "/_Standard/_Standard Information", NULL, NULL, 0, NULL  }, @selector(menu_displayStandard)},
 	};
	if( manager )
		[NSException raise: NSGenericException
		format: @"Only one instance of class 'Manager' allowed"];
	manager = [super initWithArgc: argc argv: argv ];
	if( !manager )
		[NSException raise: NSGenericException
		format: @"Fatal: Can't initialize application."];
	[manager setSignals];

	controlled_windows = [NSMutableArray new];
	kernel = [[SXKernel alloc] initWithManager: self];
	if(! controlled_windows || !kernel )
		[NSException raise: NSMallocException
			format: @"openFile: Not enough memory."];
#ifdef DEBUG
	debug_window = [TextPanel textPanelWithTitle: @"StatistX: Debug"];
#endif
	// create the main window:

/****************** GUI *****************************************/

    	top_window = [GTKWindow windowWithType:GTK_WINDOW_TOPLEVEL];
    	[top_window setBorderWidth:0];
	[top_window setTitle: @"StatistX"];
	// This is the way you set the initial window size properly
	[top_window setDefaultSize:SX_DEFAULT_WIDTH height:SX_DEFAULT_HEIGHT];
    	[top_window show]; // put something on the screen

/********** MENU ***********/
	item_factory =
	[GTKItemFactory itemFactoryWithContainerType:GTK_TYPE_MENU_BAR
			path:@"<menu>" accelGroup:nil];
	[item_factory createItems:&menus[0] count:11 forTarget:self];
	if( !item_factory ) [NSException raise: SXException
			format: @"-init: Error while creating menu bar."];

	// Call this before creation of 'Help'
	[kernel registerActionsAtManager];
	[GraphView new];
	// this creates the Help-Menu entries
	[SXHelp helpManager];

	// connect accelerators:
	[top_window addAccelGroup:[item_factory accelGroup]];

/********** TOOLBAR ************/
	handle_box = [GTKHandleBox handleBox];
	toolbar    = [GTKToolbar toolbarWithOrientation:GTK_ORIENTATION_HORIZONTAL 
				style:GTK_TOOLBAR_ICONS];
	[toolbar retain];
	[handle_box add: toolbar];

/********** NOTEBOOK ***********/

	notebook  = [GTKNotebook new];
	if(! notebook)
		[NSException raise: NSMallocException
			format: @"openFile: Not enough memory for notebook."];
	[notebook setTabPos: GTK_POS_BOTTOM];
	[notebook setScrollable: YES];

/********** TOP WINDOW ***********/

	statusbar = [GTKStatusbar statusbar];
	menubar = [item_factory getWidget:@"<menu>"];
	[menubar show];

	box = [GTKVBox vBoxWithHomogeneous:NO spacing:0];
	[box packStart:menubar    expand:NO  fill:YES padding:0];
	[box packStart:handle_box expand:NO  fill:YES padding:0];
	[box packStart:notebook   expand:YES fill:YES padding:0];
	[box packStart:statusbar  expand:NO  fill:YES padding:0];

	[top_window add: box];
    	[top_window connectSignal:@"destroy" withAction:
	    [GTKAction actionWithTarget:self selector:@selector(terminate:)]];
 
	file_list = [FileList fileListWithItemFactory: item_factory 
				maxNFiles:SX_MAX_LAST_FILES];
	[box showNow];
	lastOpenDir = [[SXSettings settings] dataDir];
	// The following method requires that the menu widgets exist. It
	// therefore has to be called this late.
	[self setUpToolbar];
	[statusbar pop: 1];
	return self;
}

-(void) setSignals {
	int i;
	for( i = 0; caught_signals[i] >= 0; i++ )
		signal( caught_signals[i], get_sig );
}

// Called during init. Sets up the previously created toolbar.
-(void) setUpToolbar {
	GTKMenuItem* item;

// Open Button
	item = (GTKMenuItem*) [item_factory getWidget:@"/File/Open..."];
	[self addToToolbarItem: item label:@"Open File" toolTip:@"Open File" 
						xpmIcon: (gchar**)ofolder_xpm];
// Print Gnuplot
	item = (GTKMenuItem*) [item_factory getWidget:@"/Gnuplot/Print Current Plot"];
	[self addToToolbarItem: item label:@"Print Gnuplot" toolTip:@"Print Gnuplot" 
						xpmIcon: (gchar**)printer_xpm];

// Standard Button
	item = (GTKMenuItem*) [item_factory getWidget:@"/Standard/Standard Information"];
	[self addToToolbarItem: item label:@"Standard" toolTip:@"Standard Information" 
						xpmIcon: (gchar**)mean_xpm];

// Histogram
	item = (GTKMenuItem*) [item_factory getWidget:@"/Standard/Histogram/Auto"];
	[self addToToolbarItem: item label:@"Histogram" toolTip:@"Histogram" 
						xpmIcon: (gchar**)histo_xpm];
// KS-Test
	item = (GTKMenuItem*) [item_factory getWidget:@"/Tests/Kolmogorov-Smirnov"];
	[self addToToolbarItem: item label:@"KS-Test" 
			toolTip:@"Kolmogorov-Smirnov-Lilliefors-Test" xpmIcon: (gchar**)normal_xpm];
// T-Test
	item = (GTKMenuItem*) [item_factory getWidget:@"/Tests/T-Test/Standard"];
	[self addToToolbarItem: item label:@"T-Test" toolTip:@"T-Test" 
						xpmIcon: (gchar**)normal2_xpm];
// Linear Regression
	item = (GTKMenuItem*) [item_factory 
		getWidget:@"/Regressions/Linear Regression/Multi Dimensional"];
	[self addToToolbarItem: item label:@"Linear Regression" toolTip:@"Linear Regression" 
						xpmIcon: (gchar**)linreg_xpm];
}


-(void) addToToolbarItem: (GTKMenuItem*) item label: (NSString*) label
		toolTip: (NSString*) tooltip_text xpmIcon: (gchar**) icon {
	NSString  * tool_tip;
	GdkBitmap * mask;
	GdkPixmap * pixmap;
	GtkWidget * top_win = [top_window gtk];
	GTKPixmap * icon_widget;
	GTKWidget * toolbar_item;
	if( !item ) return;

	pixmap = gdk_pixmap_create_from_xpm_d ( top_win->window, &mask,
            &top_win->style->white, icon );
	icon_widget = [GTKPixmap pixmapWithPixmap: pixmap mask: mask];

	toolbar_item = [toolbar appendItem: label tooltipText: tooltip_text 
			tooltipPrivateText: tooltip_text 
				icon: icon_widget	callback:NULL userData:NULL ]; 
	[toolbar_item connectSignal:@"clicked" withAction:
		[GTKAction actionWithTarget:item selector:@selector(activate)]];

	if( pixmap )
		gdk_pixmap_unref( pixmap );
}

-(void)gotSignal: (int) sig {
	char* s_signal;
	switch( sig ) {
		case SIGHUP  : s_signal = "SIGHUP"; break;
		case SIGINT  : s_signal = "SIGINT"; break;
		case SIGTERM : s_signal = "SIGTERM"; break;
		case SIGQUIT : s_signal = "SIGQUIT"; break;
		default : s_signal = "unknown signal";
	}
	if( manager )
		[manager terminate: manager];
#ifdef DEBUG
	fprintf(stderr, "terminating due to signal: %s\n", s_signal);
#endif // DEBUG
	exit(0);
}

-(void) registerDataDisplay: (DataDisplay*) dd {
	if( [dd isKindOfClass: [DataDisplay class]] 
	 && ![controlled_windows containsObject: dd] )
		[controlled_windows addObject: dd];
}

-(void) unregisterDataDisplay: (DataDisplay*) dd {
	DataSet* ds = [dd getDataSet];
	NSString *file_name = [ds getFileName];
	[file_list addFileName: file_name];
	[controlled_windows removeObject: dd];
	[[dd getDataSet] release];
	[dd release];
#ifdef DEBUG
	[debug_window appendString: [NSString stringWithFormat: 
		@"File %@ closed.\n", file_name]];
	[debug_window appendString: [NSString stringWithFormat: 
		@"Manager: remaining windows: %d\n", [controlled_windows count]]];
#endif
}
-(unsigned) countControlledDataDisplays {
	return [controlled_windows count];
}
-(void)registerAction: (NSString*) menu_path 
	target: target
	selector: (SEL) callback {
	GTKItemFactoryEntry menus[] = {
	   {{ NULL, NULL, NULL, 0, NULL }, callback},
	};
	char *buf = calloc( [menu_path length] + 1, sizeof( char ) );
	if( ! buf ) [NSException raise: NSMallocException
		format: @"registerAction: Not enough memory for menu %s.", menu_path];
	if(! item_factory ) return;
	[menu_path getCString: buf];
	menus[0].item.path = buf;
	[item_factory createItems:&menus[0] count:1 forTarget:target];
}
-(DataDisplay*)getCurrentDataDisplay {
	gint current = [notebook getCurrentPage];
	if( current < 0 || current > [controlled_windows count] - 1) return NULL;
	return (DataDisplay*) [controlled_windows objectAtIndex: (int) current];
}
-(NSArray*)getSelectedCols {
	DataDisplay* dd = [self getCurrentDataDisplay];
	return [dd getSelectedCols];
}
-(SXKernel*)getKernel {
	return kernel;
}
// return the default separators for data files
-(NSString*)getColumnSeparators {
	return @SX_COL_SEPARATORS; // allow: tab, space, comma
}

-(void)readDataFromStdin {
	DataSet* ds;
	DataDisplay* dd;

	ds = [DataSet new];
	if(! ds)
		[NSException raise: NSMallocException
			format: @"readDataFrom Stdin: Not enough memory for DataSet."];
	if( ![ds readDataFromStdin] ) {
		[Message messageWithContents: [NSString stringWithFormat: 
			@"Can't create data set for stdin."]
			title: @"StatistX: Can't read data"];
		return;
	}
	// Destroy DD as soon as posssible:
    	dd = [[DataDisplay alloc] initWithDataSet: ds];
	if(! dd)
		[NSException raise: NSMallocException
			format: @"openFile: Not enough memory for DataDisplay."];
	[self addDataDisplay: dd];
}

-(void)openFile: (NSString *) file_name {
    // load the file
	DataSet* ds = NULL;
	DataDisplay* dd = NULL;
	BOOL stop_loading = NO;

	if( !file_name || ![file_name length] ) return;
	if( ! [[NSFileManager defaultManager] 
			isReadableFileAtPath: file_name] ) {
		[Message messageWithContents: 
		[NSString stringWithFormat: @"Can't read from file %@.", file_name]
		 title: @"StatistX: Error"];
		return;
	}
	ds = [DataSet alloc];
	if(! ds)
		[NSException raise: NSMallocException
			format: @"openFile: Not enough memory for DataSet."];
	if( ![ds initWithFile: file_name] ) {
		[Message messageWithContents: [NSString stringWithFormat: 
			@"Can't open data file %s", [file_name cString]]
			title: @"StatistX: Can't open file"];
		return;
	}
	// Destroy DD as soon as posssible:
NS_DURING	
    	dd = [[DataDisplay alloc] initWithDataSet: ds];
NS_HANDLER
	if( [[localException name] isEqualToString: SXCancelOperationException] ) {
		[ds release]; ds = NULL;
		stop_loading = YES;
	} else {
		[ds release]; ds = NULL;
		[localException raise];
		stop_loading = YES;
	} 
NS_ENDHANDLER
	if( stop_loading ) return;
	if(! dd) {
		[ds release];
		[NSException raise: NSMallocException
			format: @"openFile: Not enough memory for DataDisplay."];
	}
	[self addDataDisplay: dd];
#ifdef DEBUG
	[self printOnDebug: [NSString stringWithFormat: @"File %@ opened.\n", file_name]];
#endif
}

-(void)addDataDisplay: (DataDisplay*) dd {
	GTKScrolledWindow* scroll_view;
	GtkRequisition window_size;

	if(! dd) return;
	if(! [dd isKindOfClass: [DataDisplay class]] )
		[NSException raise: NSInvalidArgumentException
			format: @"-addDataDisplay: Argument is of invalid class."];

	[self registerDataDisplay: dd];
	scroll_view = [[GTKScrolledWindow scrolledWindowWithHadjustment: NULL 
			vadjustment: NULL] retain];
	[scroll_view setPolicy: GTK_POLICY_AUTOMATIC 
			vscrollbarPolicy: GTK_POLICY_AUTOMATIC];
	[scroll_view add:dd];

	[notebook appendPage: scroll_view tabLabel: 
		[GTKLabel labelWithStr: [dd getTitle]]];
	// bring new page to front (start counting with 0):
	[notebook setPage: (gint) [self countControlledDataDisplays] - 1];

	window_size.width = SX_DEFAULT_WIDTH;
	window_size.height = SX_DEFAULT_HEIGHT;
	[scroll_view sizeRequest: &window_size];
}

/*********************************** MENU **********************************/

- (void) menu_openFile:sender {
    NSString* file_name = nil;
    GTKFileSelection *filesel =
	[[GTKFileSelection alloc] initWithTitle:@"Open File"];

    [[filesel okButton] connectSignal:@"clicked" withTarget:self
	sel:@selector(ok:)];
    [[filesel cancelButton] connectSignal:@"clicked" withTarget:self
	sel:@selector(cancel:)];
    [filesel connectSignal:@"delete_event" withTarget:self
	sel:@selector(cancel:)];
    if( lastOpenDir ) [filesel setFilename: lastOpenDir];
    [filesel hideFileopButtons];
    [filesel setPosition: GTK_WIN_POS_CENTER ];
    [filesel show];

    [statusbar push: 1 text: @" select a file ..."];


    if ([GTKApp runModalForWindow:filesel] 
    && (file_name = [filesel getFilename])) {
	[filesel destroy];
    	[filesel release];
	[statusbar pop: 1];

	[statusbar push: 2 text: 
		[NSString stringWithFormat: @" loading %@ ...", file_name]];
NS_DURING	
	[self openFile: file_name];
NS_HANDLER
	[self reportException: localException];
NS_ENDHANDLER
	// set the open panel to the last opened directory:
	if( lastOpenDir ) [lastOpenDir release];
	lastOpenDir = [[[file_name stringByDeletingLastPathComponent] 
		stringByAppendingFormat: @"%s", PATH_SEPARATOR] retain];
	[statusbar pop: 2];
	[statusbar push: 99 text: @""];
    } else {
    	[filesel destroy];
    	[filesel release];
	[statusbar pop: 1];
    }
}

- (void) ok:sender
{
    [GTKApp stopModalWithCode:YES];
}

- (void) cancel:sender
{
    [GTKApp stopModalWithCode:NO];
}

-(void) menu_closeFile: sender {
	gint current = [notebook getCurrentPage];
	if( current < 0 ) return;
	[notebook removePage: current];
	[self unregisterDataDisplay: 
		[controlled_windows objectAtIndex: (int) current]];
}
-(void)menu_selectAll {
	[[self getCurrentDataDisplay] selectAllCols];
}
-(void)menu_unselectAll {
	[[self getCurrentDataDisplay] unselectAllCols];
}

-(void)menu_displayStandard {
	gint current = [notebook getCurrentPage];
	if( current < 0 ) {
		[Message messageWithContents: @"No file open!"
			title: @"Error"];
		return;
	}
	[[controlled_windows objectAtIndex: (int) current] displayStandardData];
}

/********************************* END MENU **********************************/

/** This will open a log-file called 'path'. A previously opened log-file
   will be closed. */
-(void) setLogFileName: (NSString*) path {
	NSString * msg = [NSString stringWithFormat:@"Can't open log file\n%@", path];
	/* logfile defined in statist.h (global variable)*/
	if(logfile) fclose(logfile);
	logfile = fopen( [path cString], "w" );
	if(!logfile) {
		[Message messageWithContents: msg title: @"StatistX: Error"];
		log_set = FALSE;
		return;
	}
//fprintf(stderr, "log-file %s\n", [path cString]);
	log_set = TRUE;
}
/** Return FILE pointer to log-file. Returns NULL if no log-file open. */
-(FILE*)getLogFile {
	return logfile;
}

-(void)printOnDebug:(NSString*) debug_msg {
	if( debug_msg && [debug_msg isKindOfClass: [NSString class]] ) 
		[debug_window appendString: debug_msg];
}

-(void)reportException: (NSException*) exception {
	NSString* msg = [NSString stringWithFormat: 
		@"Internal error\nReason:\n%@", [exception reason]];
	[self printOnDebug: msg];
	[Message messageWithContents: msg title: @"StatistX: Internal error"];
}

-(void)terminate: sender {
	if( terminated ) return;
#ifdef DEBUG
	fprintf(stderr, "Application will terminate ...\n");
#endif
	terminated = YES;
	finish_statist();
	[super terminate: sender];
}

-(void)dealloc {
 	if( !terminated )
		[self terminate: self];
	[toolbar release];
	[controlled_windows release];
	[notebook release];
	[kernel release];
	manager = nil;
#ifdef DEBUG
	fprintf(stderr, "Manager deallocated\n");
#endif // DEBUG
	return [super dealloc];
}
@end

#undef IN_MANAGER_M  /* This should be the last line. */

