/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 2000
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/* SXHelp.h
 * Administrates help, help-browsers, etc.
 *
 */
#include "SXHelp.h"


#include <Foundation/NSException.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSDictionary.h>
#include <GToolKit/GToolKit.h>

#include "HTMLView.h"
#include "Manager.h"
#include "Message.h"
#include "SXSettings.h"

#include "sx_defs.h"

static SXHelp * help_manager = NULL;

@implementation SXHelp
+helpManager {
	if( !help_manager ) help_manager = [self new];
	return help_manager;
}
-init {
	SXSettings * settings = [SXSettings settings];
	NSString * index_file;
	if( help_manager )
		[NSException raise: NSGenericException
			format: @"-init: illegal call of init"];
	help_manager = [super init];

	base_dir = [[settings valueForKey: @"HELP_DIR"] stringByExpandingTildeInPath];
	if( ![base_dir length] ) {
		base_dir = [@SX_DOC_DIR stringByExpandingTildeInPath];
	}
	[base_dir retain];
	// load help index
	index_file = [base_dir stringByAppendingPathComponent: @"help.idx"];	
	if( ![self loadIndexFile: index_file] ) {
		help_available = NO;
		[Message messageWithContents: @"Can't initialize help. No on-line help available."
			title: @"StatistX: Error"];
	} else {
		[[Manager getManager] registerAction: @"/_Help/StatistX _Help"
			target: self selector: @selector(raiseMainDialog)];
	}
	[[Manager getManager] registerAction: @"/_Help/_About StatistX"
		target: self selector: @selector(raiseAboutDialog)];
	return self;
}

-(BOOL)loadIndexFile: (NSString*) full_path {
	NSMutableDictionary * dict;
	NSString * index_file = [NSString stringWithContentsOfFile: full_path];
//fprintf(stderr, "looking for %s\n", [full_path cString]);
	dict = [index_file propertyListFromStringsFileFormat];
	if( ![dict count] ) return NO;
	if( index ) [index addEntriesFromDictionary: dict];
	else 	{ // no existing index, create a new one
		index = [NSMutableDictionary dictionaryWithDictionary: dict];
		[index retain];
	}
	return (index != NULL);
}

-(BOOL)loadHelpFile: (NSString*) full_path {
	NSString * file_cont = [NSString stringWithContentsOfFile: full_path];
	int length = [file_cont length];
	if( !length ) return NO;
	[help_browser clear]; // clear text 
	[help_browser freeze];
	[help_browser appendHTMLCode: file_cont];
	[help_browser appendText: @"\n"]; // looks usually better
	[help_browser thaw];
	return YES;
}
/** Add text to the index, 'help_text' will be shown when 'key' is selected */
-(void)addToIndex: (NSString*) key path: (NSString*) a_help_file {
	[index setObject: a_help_file forKey: key];
}
-(void)listActivate:sender :(gint)row :(gint)col {
	NSString * text = NULL;
	NSString * title;
	if( !index_list ) return;
	[index_list getText:row column: col text: &text]; 
	if( ![text length] ) {
		title = [NSString stringWithFormat: 
				@"Sorry, no help found for \"%@\".", text];
		[Message messageWithContents: title	title: @"StatistX Help: Error"];
		return;
	}
	title = [NSString stringWithFormat: @"StatistX Help: %@", text];
	[self showHelpForIndex: text];
	[window setTitle: title];
}
/** Open the main help dialog (index) */
-(void)raiseMainDialog {
	GTKButton*  ok = [GTKButton buttonWithLabel: @"OK"];
	GTKVBox*    v_box;
	GTKHBox*    h_box;
	GtkButton  *gtk_button = [ok gtk];
	NSArray *	 allKeys = [index allKeys];
	NSArray *	 row;
	NSString	*startup_help__file;
	unsigned	 i, cnt = [allKeys count];
	GTKHPaned *	 paned = [GTKHPaned hPaned];
	GTKScrolledWindow * scroll_area
		= [GTKScrolledWindow scrolledWindowWithHadjustment: NULL vadjustment: NULL]; 
	if( window ) {
		[window show];
		return;
	}
	index_list	 = [GTKCList cListWithColumns: 1];
	help_browser = [HTMLView textWithHadj:NULL vadj:NULL];
	[help_browser setEditable: FALSE];
	[scroll_area add: help_browser];
	[scroll_area setPolicy: GTK_POLICY_AUTOMATIC 
		vscrollbarPolicy:  GTK_POLICY_ALWAYS];

	[paned add1: index_list];
	[paned add2: scroll_area];

	for( i = 0; i < cnt; i++ ) {
	/* Create one array for each row. Each row has just
	   one column, each array thus contains only one element. */
		id entry = [allKeys objectAtIndex: i];
		row = [NSArray arrayWithObjects:&entry count: 1];
		[index_list append:row];
	}
	[index_list sort];
	[index_list setSelectionMode: GTK_SELECTION_SINGLE];
	[index_list connectSignal:@"select_row" 
		withTarget:self sel: @selector(listActivate:::) ];
	[index_list setUsize:100 height:80];
	[paned show];
	window = [GTKDialog dialog];

	v_box = (GTKVBox*) [window vbox];
	h_box = (GTKHBox*) [window actionArea];
	
	[ok setUsize: 85 height: 35];
	[h_box packStart:ok expand:YES fill:NO padding:0];
	gtk_misc_set_padding((GtkMisc *) gtk_button->child, 4, 0);
      GTK_WIDGET_SET_FLAGS(gtk_button, GTK_CAN_DEFAULT);
	[ok connectSignal:@"clicked" withAction:
        	[GTKAction actionWithTarget:window selector:@selector(destroy)]];
	[ok grabDefault];
	[ok grabFocus];

	[v_box packStart:paned expand:YES fill:YES padding:0];

	[window setUsize: 200 height: 100];
	[window connectSignal: @"destroy" withTarget: self sel: @selector(windowClosed)];
	[window setDefaultSize:600 height:450];
	[window setTitle: @"StatistX Help"];
	[window setPosition: GTK_WIN_POS_CENTER ];
	[window show];

	[window retain];
	[index_list retain];
	[help_browser retain];

	// load initial help into the browser
	startup_help__file = [base_dir stringByAppendingPathComponent: @"index.html"];
	[self loadHelpFile: startup_help__file];
}
// open an "About StatistX" dialog
-(void)raiseAboutDialog {
	GTKLabel* label;
	NSString* text;
	text = [NSString stringWithFormat: about_StatistX,
	[Manager version]];
	label = [GTKLabel labelWithStr: text];
	[label setPadding: 5 ypad: 10];
	[OKDialog OKDialogWithContents: label title: @"About StatistX"];
}
-(void)windowClosed {
	[window release]; 
	window = NULL;
	[index_list release]; 
	index_list = NULL;
	[help_browser release]; 
	help_browser = NULL;
}
/** Show help for term "index", returns NO if no help available */
-(BOOL)showHelpForIndex: (NSString*) index_entry {
	NSString * file_name = [[index objectForKey: index_entry] stringByExpandingTildeInPath];
	if( ![file_name length] ) return NO;
	if( ![file_name isAbsolutePath] ) {
		file_name = [base_dir stringByAppendingPathComponent: file_name]; 
	}
	return [self loadHelpFile: file_name];
}

-(void) dealloc {
	[window release]; 
	[index_list release]; 
	[help_browser release]; 
	[base_dir release];
	[super dealloc];
}
@end
