/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 1999 - 2000
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/**	SXKernel.m - perform statistical analysis.
 *
 *	Every statistics function is implemented twice:
 *	One provides the interface to statist (i.e. wraps the c-function)
 *	and the second obtains the necessary parameters (i.e. asks the
 *	user for parameters and calls the first method).
 *
 *  Since Gnuplot currently does not support locale dependent
 *  parsing of numbers, LC_NUMERIC is set to the default C-locale
 *  before calling a statist-funnction.
 *
 */
#include <string.h>
#include <locale.h>

#include <Foundation/NSException.h>
#include <Foundation/NSString.h>

#include "SXKernel.h"
#include "Column.h"
#include "Message.h"
#include "SelectColumnDialog.h"
#include "HistogramDialog.h"
#include "SXException.h"

#include "statist.h"	// statist header files

#define SX_SET_LOCALE	\
	SX_old_locale = setlocale (LC_NUMERIC, NULL); \
	SX_saved_locale = strdup (SX_old_locale); \
	if (SX_saved_locale == NULL) { \
		[NSException raise: NSMallocException \
			format: @"set locale: Not enough memory."]; \
	} \
	setlocale (LC_ALL, "C"); 

#define SX_RESET_LOCALE \
	setlocale (LC_ALL, SX_saved_locale); \
	free(SX_saved_locale); SX_saved_locale = NULL;


@implementation SXKernel
+kernelWithManager: (Manager*) m {
	return [[[self alloc] autorelease] initWithManager: m];
}

-initWithManager: (Manager*) m {
	[super init];
	manager = [m retain];
	return self;
}

// register all known statistical methods at the given Manager
-(void)registerActionsAtManager {
//==================== Standard Menu =======================================
	[manager   registerAction: @"/_Standard/_Histogram/_Auto"
				target: self
				selector: @selector(histogramAuto:)];
	[manager   registerAction: @"/_Standard/_Histogram/_User Defined"
				target: self
				selector: @selector(histogramUserDefined:)];
	[manager   registerAction: @"/Standard/_Outlier"
				target: self
				selector: @selector(outlier:)];

//==================== Regression & Correlation Menu =======================
	[manager   registerAction: @"/_Regressions/_Rank Order"
				target: self
				selector: @selector(rankOrder:)];
	[manager   registerAction: @"/Regressions/_Linear Regression/_One Dimensional"
				target: self
				selector: @selector(linearRegression:)];
	[manager   registerAction: @"/Regressions/Linear Regression/_Multi Dimensional"
				target: self
				selector: @selector(multipleRegression:)];
	[manager   registerAction: @"/Regressions/Partial L_in. Correlation"
				target: self
				selector: @selector(partLinearCorr:)];
	[manager   registerAction: @"/Regressions/_Polynomial Regression"
				target: self
				selector: @selector(polynomialRegresssion:)];
	[manager   registerAction: @"/Regressions/_Correlation Matrix"
				target: self
				selector: @selector(correlationMatrix:)];
	[manager   registerAction: @"/Regressions/_Spearmann Rank Correlation"
				target: self
				selector: @selector(rankMatrix:)];
	[manager   registerAction: @"/Regressions/Cross _Validation"
				target: self
				selector: @selector(crossValidate:)];
	[manager   registerAction: @"/Regressions/Random _Tupel"
				target: self
				selector: @selector(randomTupel:)];
	[manager   registerAction: @"/Regressions/Point _Biserial Regr."
				target: self
				selector: @selector(pointBiserialRegression:)];
	[manager   registerAction: @"/Regressions/_Probit Analysis"
				target: self
				selector: @selector(probitAnalysis:)];
	[manager   registerAction: @"/Regressions/Probit %"
				target: self
				selector: @selector(probitAnalysisNoCountDose:)];
//==================== Tests Menu ===========================================
	[manager   registerAction: @"/_Tests/_T-Test/_Standard"
				target: self
				selector: @selector(tTest:)];
	[manager   registerAction: @"/Tests/_T-Test/_Paired"
				target: self
				selector: @selector(tTest:)];
	[manager   registerAction: @"/Tests/_Kolmogorov-Smirnov"
				target: self
				selector: @selector(ksTest:)];
	[manager   registerAction: @"/Tests/_Chi^2 (4 Fields) Test"
				target: self
				selector: @selector(fourFieldsTest:)];
	[manager   registerAction: @"/Tests/_2 Items-Test"
				target: self
				selector: @selector(twoItemsTest:)];
	[manager   registerAction: @"/Tests/_U-Test"
				target: self
				selector: @selector(uTest:)];
	[manager   registerAction: @"/Tests/K_ruskal-Wallis"
				target: self
				selector: @selector(hTest:)];
	[manager   registerAction: @"/Tests/_Wilcoxon-Test"
				target: self
				selector: @selector(wilcoxonTest:)];
	[manager   registerAction: @"/Tests/_Equal Frequency/_One Set"
				target: self
				selector: @selector(equalFrequTest:)];
	[manager   registerAction: @"/Tests/_Equal Frequency/_Comparing Meas.-Exp."
				target: self
				selector: @selector(compareFrequTest:)];
}

/*  Get exactly number_of_cols
 *  columns from the manager.
 *  Creates error message otherwise.
 *  This method returns NULL if too few or too many collumns are selected.
 */
-(NSArray*)getCols: (unsigned) number_of_cols {
	NSArray* sel_cols = [manager getSelectedCols];
	if( [sel_cols count] != number_of_cols ) {
		NSString* msg = [NSString stringWithFormat: 
			@"Exactly %d columns have to be\n selected for this operation.", 
			number_of_cols];
		if( number_of_cols == 1 )
			msg = @"Exactly one column has to be\n selected for this operation.";
		[Message messageWithContents: msg title: @"Error"];
		return NULL;
	}
	return sel_cols;
}

// get at least one column from the manager, error, if no columns
-(NSArray*)getNCols {
	NSArray* sel_cols = [manager getSelectedCols];
	if( [sel_cols count] < 1 ) {
			[Message messageWithContents: @"No columns selected!"
			title: @"Error"];
		return NULL;
	}
	return sel_cols;
}

//==========================================================================
//>>>>>> statistical stuff:

//==================== Standard Menu =======================================
-(void) histogramAuto: sender {
	NSArray* cols = [self getCols: 1];
	if( !cols ) return;
	[self histogramForCol: (Column*) [cols objectAtIndex: 0]
		n_classes: 0 max: 0 min: 0];
}
-(void) histogramUserDefined: sender {
	NSArray* cols = [self getCols: 1];
	Column* col;
	HistogramDialog* dialog;
	BOOL result;
	int n_cl;
	double max, min;
	if( !cols ) return;
	col = (Column*) [cols objectAtIndex: 0];
	dialog = [[HistogramDialog alloc] initWithColumn: col];
	if( ! dialog ) 	[NSException raise: NSGenericException
		format: @"-histogramUserDefined: Can't create HistogramDialog"];
	result = [GTKApp runModalForWindow: dialog];
	[dialog destroy];
	if( !result ) { [dialog release]; return; }
	n_cl = [dialog getNClasses];
	max = [dialog getMax];
	min = [dialog getMin];
	[dialog release];
	[self histogramForCol: col n_classes: n_cl max: max min: min];
}
-(void) histogramForCol: (Column*) col 
		  n_classes: (int) n_cl // if zero -> auto determination of classes
		  	  max: (double) max
			  min: (double) min {
	if( !col ) return;
	if( n_cl == 0 ) { // auto
		min = [col min];
		max = [col max];
	}
	if( min == max ) {
		[Message messageWithContents:
			@"Can't create histogram:\n\
			  all numbers are equal."
		   title: @"Histogram: Error"];
		return;
	}
	if( min > max )
		[NSException raise: NSInvalidArgumentException 
			format: @"-histogramForCol: min > max"];


SX_SET_LOCALE
	/* calc and plot the histogram	*/
	histogram([col getArray], [col count],n_cl,min,max);
SX_RESET_LOCALE
}
-(void) probitAnalysis: sender {
	Column *dose, *effect, *cnt;
	NSArray* t_cols = [self getCols: 3];
	NSMutableArray* cols;
	if( !t_cols ) return;
	cols = [NSMutableArray arrayWithArray: t_cols];
	if( !cols ) [NSException raise: NSMallocException
			format: @"-probitAnalysis: not enough memory"];
	dose = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column which\ncontains the doses."
		title: @"Probit Analysis: Select Column, Dose"];
	if( !dose ) return;
	[cols removeObject: dose];
	cnt = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column which contains the\n\
				numbers of individuals for each dose.\n\
				(Must contain positive integers only.)"
		title: @"Probit Analysis: Select Column, Numbers"];
	if( !cnt ) return;
	while( ![cnt hasPosIntegers] ) {
		[Message messageWithContents:
			@"Column with numbers of individuals must contain\n\
			  positive integers only."
		   title: @"Probit Analysis: Error"];
		cnt = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column which contains the\n\
				numbers of individuals for each dose.\n\
				(Must contain positive integers only.)"
		title: @"Probit Analysis: Select Column, Numbers"];
		if( !cnt ) return;
	}
	[cols removeObject: cnt];
	effect = [cols objectAtIndex: 0];
	[self probitAnalysisDose: dose count: cnt effect: effect];
}
-(void) probitAnalysisDose: (Column*) c_1 
			   count: (Column*) c_2 
			  effect: (Column*) c_3 {
	unsigned n = [c_1 count];
	double *dose, *effect, *cnt;
	dose   = [c_1 getArray];
	cnt  = [c_2 getArray];
	effect = [c_3 getArray];
	if( dose == NULL || dose == NULL || cnt == NULL )
		[NSException raise: NSInvalidArgumentException 
			format: @"-probitAnalaysis: got empty columns"];
	if( [c_2 count] != n || [c_3 count] !=n ) {
		[Message messageWithContents:
				@"Columns don't have equal lengths."
			   title: @"Probit analysis: Error"];
		return;
	}
SX_SET_LOCALE
	probit(dose, cnt, effect, n);
SX_RESET_LOCALE
}
-(void) probitAnalysisNoCountDose: sender {
	Column *dose, *effect;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	dose   = (Column*)[cols objectAtIndex: 0];
	effect = (Column*)[cols objectAtIndex: 1];
	[self probitAnalysisDose: dose effect: effect];
}
-(void) probitAnalysisDose: (Column*) c_1 
			  effect: (Column*) c_2 {
	unsigned i, n;
	double *dose, *effect, *cnt;
	dose   = [c_1 getArray];
	effect = [c_2 getArray];
	if( dose == effect )
		[NSException raise: NSInvalidArgumentException 
			format: @"-probitAnalaysis: got equal columns"];
	n = [c_1 count];
	cnt    = calloc( n, sizeof(double) );
	if( !cnt ) [NSException raise: NSMallocException
			format: @"-probitAnalysis: not enough memory"];
	for( i = 0; i < n; i++ ) cnt[i] = 100;
SX_SET_LOCALE
	probit(dose, cnt, effect, n);
SX_RESET_LOCALE
}

-(void) outlier: sender {
	NSArray* cols = [self getCols: 1];
	Column* c = (Column*) [cols objectAtIndex: 0];
	if( !c ) return;
	[self outlierOfCol: c];
}
-(void) outlierOfCol: (Column*) c {
	unsigned n;
	n = [c count];
	if( !n ) return; // Dont blame the user, just ignore it.
SX_SET_LOCALE
	outlier( [c getArray], n );
SX_RESET_LOCALE
}
//==================== Regression & Correlation Menu =======================

-(void) rankOrder: sender {
	NSArray* cols = [self getCols: 2];
	Column* c1, *c2;
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	[self rankOrderForCols: c1 : c2];
}
-(void) rankOrderForCols: (Column*) c1 : (Column*) c2 {
	unsigned n1 = [c1 count], n2 = [c2 count];
	if( n1 == 0 || n2 == 0 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-rankOrderForCols: empty column"];
	if( n1 != n2 ) {
		[Message messageWithContents:
				@"Columns don't have equal lengths."
			   title: @"Rank order: Error"];
		return;
	}

SX_SET_LOCALE
	rank_order([c1 getArray], [c2 getArray], n1);
SX_RESET_LOCALE
}
-(void) linearRegression: sender {
	NSArray* cols = [self getCols: 2];
	Column *c1, *c2, *temp;
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	temp = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column that\ncontains the dependent factor."
		title: @"Linear Regression: Select Column"];
	if( !temp ) return;
	if( temp != c1 ) c2 = c1, c1 = temp; // flip the columns
	[self linearRegressionForCols: c1 : c2];
}
-(void) linearRegressionForCols: (Column*) c1 : (Column*) c2 {
	if( [c1 count] == 0 || [c2 count] == 0 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-linearRegressionForCols: empty column"];
SX_SET_LOCALE
	lin_reg([c2 getArray], [c1 getArray], [c1 count]);
SX_RESET_LOCALE
}
-(void) multipleRegression: sender {
	Column *treatment; // independant factor
	unsigned col_n;
	NSArray* user_cols = [self getNCols];
	NSMutableArray* cols;
	if( !user_cols ) return;
	col_n = [user_cols count];
	if( col_n < 2 ) {
		[Message messageWithContents:
				@"A regression requires at\nleast two columns."
			   title: @"Multiple regression: Error"];
		return;
	}
	cols = [NSMutableArray arrayWithArray: user_cols];
	if( !cols ) [NSException raise: NSMallocException
			format: @"-multipleRegression: not enough memory"];
	treatment = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column \ncontaining the dependent factor."
		title: @"Linear Regression: Select Column"];
	if( !treatment ) return;
	// drop the treatment column
//make_new_col("resi", 6);
//fprintf(stderr, "cols contains %d objects\n", [cols count]);
	[cols removeObject: treatment];
//fprintf(stderr, "cols contains %d objects\n", [cols count]);
	[self multipleRegressionTreatment: treatment dependent: cols];
}
-(void) multipleRegressionTreatment: (Column*) treat dependent: (NSArray*) cols {
	double **columns;
	double* treatment = [treat getArray];
	unsigned i, col_n, row_n = [treat count];
	col_n = [cols count];
	if( !cols || col_n == 0 ) return;
	columns    = calloc( col_n, sizeof(*columns) );
	if( !columns ) [NSException raise: NSMallocException
			format: @"-multipleRegressionForCols: not enough memory"];
	for( i = 0; i < col_n; i++ ) {
		columns[i] = [(Column*)[cols objectAtIndex: i] getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-multipleRegressionForCols: can't get column number %d", i];
	}
SX_SET_LOCALE
	multiple_reg(treatment, columns, row_n, col_n);
	free( columns );
SX_RESET_LOCALE
}
-(void) partLinearCorr: sender {
	unsigned col_n;
	NSArray* cols = [self getNCols];
	if( !cols ) return;
	col_n = [cols count];
	if( col_n < 3 || col_n > 5 ) {
		[Message messageWithContents:
				@"Partial lin. correlation\nrequires 3, 4 or 5 columns."
			   title: @"Partial linear correlation: Error"];
		return;	
	}
	[self partLinearCorrForCols: cols];
}
-(void) partLinearCorrForCols: (NSArray*) cols {
	double **columns;
	unsigned i, col_n, length;
	col_n = [cols count];
	if( col_n < 3 || col_n > 5 ) {
		[NSException raise: NSInvalidArgumentException 
			format: @"- partLinearCorrForCols: number of columns has\
			to be 3,4 or 5. #cols: %d", col_n];
	}
	columns    = calloc( col_n, sizeof(*columns) );
	if( !columns ) [NSException raise: NSMallocException
			format: @"-partLinearCorrForCols: not enough memory"];
	length = [(Column*)[cols objectAtIndex: 0] count];
	for( i = 0; i < col_n; i++ ) {
		columns[i] = [(Column*)[cols objectAtIndex: i] getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-partLinearCorrForCols: can't get column number %d",i];
	}
SX_SET_LOCALE
	// col_n has to be 3,4 or 5!
	part_corr(columns, length, col_n);
	free( columns );
SX_RESET_LOCALE
}

-(void) polynomialRegresssion: sender {
	unsigned order;
	Column *c1, *c2, *temp;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	order = 2; // should ask the user!
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	temp = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column that\ncontains the dependent factor."
		title: @"Polynomial Regression: Select Column"];
	if( ! temp ) return;
	if( temp != c1 ) c2 = c1, c1 = temp; // flip the columns
	[self polynomialRegresssionWithOrder: order : c1 : c2];
}
-(void) polynomialRegresssionWithOrder: (unsigned) order : (Column*) col_1 : (Column*) col_2 {
	double *c1, *c2;
	unsigned length;
	length = [col_1 count];
	if( order <= 0 || order > MPOLY )
		[NSException raise: NSInvalidArgumentException 
			format: @"-polyRegresssionForCols: invalid order"];
	c1 = [col_1 getArray];
	c2 = [col_2 getArray];
	if( c1 == c2 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-polyRegresssionForCols: equal columns"];
SX_SET_LOCALE
	poly_reg(c2, c1, length, order);
SX_RESET_LOCALE
}

-(void) correlationMatrix: sender {
	unsigned col_n;
	NSArray* cols = [self getNCols];
	if( !cols ) return;
	col_n = [cols count];
	if( col_n < 2 ) {
		[Message messageWithContents:
				@"Correlation requires at \nleast two columns."
			   title: @"Correlation Matrix: Error"];
		return;	
	}
	[self corrMatrixForCols: cols];
}
-(void) corrMatrixForCols: (NSArray*) cols {
	double **columns;
	unsigned i, length, col_n;
	if( !cols ) return;
	col_n = [cols count];
	columns = calloc( col_n, sizeof(*columns) );
	if( !columns ) [NSException raise: NSMallocException
			format: @"-corrMatrixForCols: not enough memory"];
	length = [(Column*)[cols objectAtIndex: 0] count];
	for( i = 0; i < col_n; i++ ) {
		columns[i] = [(Column*)[cols objectAtIndex: i] getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-corrMatrixForCols: can't get column number %d",i];
	}
SX_SET_LOCALE
	correl_matrix(columns, length, col_n);
	free( columns );
SX_RESET_LOCALE
}
// Spearmann rank-correlation
-(void) rankMatrix: sender {
	unsigned col_n;
	NSArray* cols = [self getNCols];
	if( !cols ) return;
	col_n = [cols count];
	if( col_n < 2 ) {
		[Message messageWithContents:
				@"Rank correlation requires at\nleast two columns."
			   title: @"Rank correlation matrix: Error"];
		return;	
	}
	[self rankMatrixForCols: cols];
}
-(void) rankMatrixForCols: (NSArray*) cols {
	double **columns;
	unsigned i, length, col_n;
	if( !cols ) return;
	col_n = [cols count];
	if( !col_n ) return;
	columns = calloc( col_n, sizeof(*columns) );
	if( !columns ) [NSException raise: NSMallocException
			format: @"-rankMatrixForCols: not enough memory"];
	length = [(Column*)[cols objectAtIndex: 0] count];
	for( i = 0; i < col_n; i++ ) {
		columns[i] = [(Column*)[cols objectAtIndex: i] getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-rankMatrixForCols: can't get column number %d",i];
	}
SX_SET_LOCALE
	rank_matrix(columns, length, col_n);
	free( columns );
SX_RESET_LOCALE
}

-(void) crossValidate: sender {
	Column *treatment; // independant factor
	unsigned col_n;
	NSMutableArray* cols = [self getNCols];
	if( !cols ) return;
	cols = [NSMutableArray arrayWithArray: cols];
	col_n = [cols count];
	if( col_n < 2 ) {
		[Message messageWithContents:
				@"Cross validation requires at\nleast two columns."
			   title: @"Cross validation: Error"];
		return;
	}
	treatment = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column that\ncontains the dependent factor."
		title: @"Cross Validation: Select Column"];
	if( ! treatment ) return;
	// drop the treatment column
	[cols removeObject: treatment];
	[self crossValidateIndependent: treatment dependent: cols];
}
-(void) crossValidateIndependent: (Column*) indep dependent: (NSArray*) cols {
	double **columns;
	unsigned i, col_n;
	if( !cols || !indep ) return;
	col_n = [cols count];
	if( !col_n ) return;
	columns    = calloc( col_n, sizeof(*columns) );
	if( !columns ) [NSException raise: NSMallocException
			format: @"-crossValidate: not enough memory"];
	for( i = 0; i < col_n; i++ ) {
		columns[i] = [(Column*)[cols objectAtIndex: i] getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-crossValidate: can't get column number %d", i];
	}
SX_SET_LOCALE
	cross_validate([indep getArray], columns, [indep count], col_n);
	free( columns );
SX_RESET_LOCALE
}

-(void) randomTupel: sender {
	Column *indep; // independant factor
	unsigned number_of_tupels, col_n;
	NSMutableArray* cols = [self getNCols];
	if( !cols ) return;
	cols = [NSMutableArray arrayWithArray: cols];
	col_n = [cols count];
	if( col_n < 2 ) {
		[Message messageWithContents:
				@"Randomization requires at \nleast two columns."
			   title: @"Random tupel: Error"];
		return;
	}
	number_of_tupels = 3; // should ask user!
	indep = [SelectColumnDialog runDialogWithColumns: cols 
		message:  @"Select the column that\ncontains the dependent factor."
		title: @"Cross Validation: Select Column"];
	if( ! indep ) return;
	// drop the independent column
	[cols removeObject: indep];
	[self randomTupelWithNTupels: number_of_tupels
				independent: indep 
				  dependent: cols];
}
-(void) randomTupelWithNTupels: (unsigned) number_of_tupels
			 independent: (Column*) indep 
			   dependent: (NSArray*) cols {
	double **columns;
	unsigned i, col_n;
	if( !cols || !indep ) return;
	col_n = [cols count]; if( !col_n ) return;
	columns    = calloc( col_n, sizeof(*columns) );
	if( !columns ) [NSException raise: NSMallocException
			format: @"-randomTupelForCols: not enough memory"];
	for( i = 0; i < col_n; i++ ) {
		columns[i] = [(Column*)[cols objectAtIndex: i] getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-randomTupelWithNTupels: can't get column number %d", i];
	}
SX_SET_LOCALE
	random_tupel([indep getArray], columns, [indep count], col_n, number_of_tupels);
	free( columns );
SX_RESET_LOCALE
}

// First column must contain only 1's or 0's.
-(void) pointBiserialRegression: sender {
	NSArray* cols = [self getCols: 2];
	Column* c1, *c2, *temp;
	if( !cols ) return;
	c1 = (Column*) [cols objectAtIndex: 0];
	c2 = (Column*) [cols objectAtIndex: 1];
	if( ! [c1 isBinary] ) { // flip
	    temp = c1; c1 = c2; c2 = temp;
	    if( ! [c1 isBinary] ) {
		[Message messageWithContents:
		@"Either of the columns must\ncontain 0s or 1s only!"
			title: @"Point biserial regression: Error"];
		return;
	    }
	}
	[self pointBiserialRegForCols: c1 : c2];
}
-(void) pointBiserialRegForCols: (Column*) col_1 : (Column*) col_2 {
	double* c1;
	double* c2;
	unsigned n;
	if( !col_1 || !col_2 ) return;
	c1 = [col_1 getArray];
	c2 = [col_2 getArray];
	if( !c1 || !c2 )
		[NSException raise: SXEmptyColumnException
		format: @"-pointBiserialRegForCols: column doesn't exist" ];
	if( c1 == c2 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-pointBiserialRegForCols: got equal columns"];
	n = [col_1 count];
	if( ! [col_1 isBinary] ) {
		[NSException raise: NSInvalidArgumentException 
			format: @"-pointBiserialRegForCols: Column %s must \
			contain 0s or 1s only!", [col_1 getLabel]];
	}
SX_SET_LOCALE
	point_biserial_reg(c1, c2, n);
SX_RESET_LOCALE
}

//==================== Tests Menu ===========================================
// t-Test zum Vergleich zweier Mittelwerte aus Stichproben
-(void) tTest: sender {
	Column* c1, *c2;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	[self tTestForCols: c1 : c2];
}
-(void) tTestForCols: (Column*) c1 : (Column*) c2 {
	if( !c1 || !c2 )
		[NSException raise: SXEmptyColumnException
		format: @"-tTestForCols: Null pointer as argument." ];
	if( c1 == c2 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-tTestForCols: got equal columns"];
SX_SET_LOCALE
	t_test([c1 getArray], [c1 count], [c2 getArray], [c2 count]);
SX_RESET_LOCALE
}

// t-test zum Vergleich paarweise erhobener Stichproben
-(void) pairTTest: sender {
	Column* c1, *c2;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	[self pairTTestForCols: c1 : c2];
}
-(void) pairTTestForCols: (Column*) c1 : (Column*) c2 {
	if( !c1 || !c2 )
		[NSException raise: SXEmptyColumnException
		format: @"-pairTTestForCols: column doesn't exist" ];
	if( c1 == c2 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-pairTTestForCols: got equal columns"];
SX_SET_LOCALE
	pair_t_test([c1 getArray], [c2 getArray], [c2 count]);
SX_RESET_LOCALE
}

// Test auf Normalverteilung (KS-Lillifors-Test)
-(void) ksTest: sender {
	NSArray* cols = [self getCols: 1];
	if( !cols ) return;
	[self ksTestForCol: (Column*)[cols objectAtIndex: 0]];
}
-(void) ksTestForCol: (Column*) col {
	if( !col ) return;
SX_SET_LOCALE
	kolmo_test([col getArray], [col count]);
SX_RESET_LOCALE
}
// Chi^2-Vierfeldertafel
// Both columns must contain only 1's or 0's
-(void) fourFieldsTest: sender {
	Column* c1, *c2;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	if( ![c1 isBinary] || ![c2 isBinary] ) {
		[Message messageWithContents: @"Both columns must contain 0s or 1s only!"
			title: @"Chi^2 (Four Fields Test): Error"];
		return;
	}
	[self fourFieldsTestForCols: c1 : c2];
}
// Chi^2-Vierfeldertafel
-(void) fourFieldsTestForCols: (Column*) c1 : (Column*) c2 {
	if( !c1 || !c2 ) return;
	if( c1 == c2 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-fourFieldsTest: got equal columns"];
SX_SET_LOCALE
	vierfeld_test([c1 getArray], [c2 getArray], [c2 count]);
SX_RESET_LOCALE
}

// CHI^2 Zwei-Merkmale-Tafeltest
-(void) twoItemsTest: sender {
	NSArray* cols = [self getNCols];
	if( !cols ) return;
	[self twoItemsTestForCols: cols];
}
-(void) twoItemsTestForCols: (NSArray*) cols {
	double **columns;
	unsigned i, length, col_n;
	if( !cols ) return;
	col_n = [cols count];
	for( i = 0; i < col_n; i++ ) {
		if(! [(Column*)[cols objectAtIndex: i] allEntriesLargerThan: 0] ){
			[Message messageWithContents: @"All columns must contain only positive numbers!"
				title: @"Two Items Test: Error"];
			return;			
		}
	}
	columns = calloc( col_n, sizeof(*columns) );
	if( !columns ) [NSException raise: NSMallocException
			format: @"-twoItemsTestForCols: not enough memory"];
	length = [(Column*)[cols objectAtIndex: 0] count];
	for( i = 0; i < col_n; i++ ) {
		columns[i] = [(Column*)[cols objectAtIndex: i] getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-twoItemsTestForCols: can't get column number %d",i];
	}
SX_SET_LOCALE
	tafel_test(columns, length, col_n);
	free( columns );
SX_RESET_LOCALE
}

// u-Test (Test auf Unabhaengigkeit 2er Stichproben)
-(void) uTest: sender {
	Column* c1, *c2;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	[self uTestForCols: c1 : c2];
}
-(void) uTestForCols: (Column*) c1 : (Column*) c2 {
	if( !c1 || !c2 ) return;
	if( c1 == c2 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-uTestForCols: got equal columns"];
SX_SET_LOCALE
	u_test([c1 getArray], [c1 count], [c2 getArray], [c2 count]);
SX_RESET_LOCALE
}

// H-Test (Kruskal-Wallis) fuer k unabhaengige Stichproben
// (MIND. 3 SPALTEN)
-(void) hTest: sender {
	unsigned col_n;
	NSArray* cols = [self getNCols];
	if( !cols ) return;
	col_n = [cols count];
	if( col_n < 3 ) {
		[Message messageWithContents:
			@"At least 3 columns required for \nKruskal-Wallis test (H-Test)."
			title: @"H-Test: Error"];
		return;
	}
	[self hTestForCols: cols];
}
-(void) hTestForCols: (NSArray*) cols {
	double **columns;
	unsigned col_n, i, *lengths;
	if( !cols ) return;
	col_n = [cols count];
	columns = calloc( col_n, sizeof(*columns) );
	lengths = calloc( col_n, sizeof(*lengths) );
	if( !columns || !lengths ) 
		[NSException raise: NSMallocException
			format: @"-hTest: not enough memory"];
	for( i = 0; i < col_n; i++ ) {
		Column* t_col = (Column*)[cols objectAtIndex: i];
		columns[i] = [t_col getArray];
		if( !columns[i] )
			[NSException raise: SXEmptyColumnException
			format: 
			@"-hTest: can't get column number %d",i];
		lengths[i] = [t_col count];
	}
SX_SET_LOCALE
	// requires at least 3 cols!
	kruskal_test(columns, lengths, col_n);
SX_RESET_LOCALE
}
// Wilcoxon-Rang-Test fuer paarweise erhobene Stichproben
-(void) wilcoxonTest: sender {
	Column* c1, *c2;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	[self wilcoxonTestForCols: c1 : c2];
}
-(void) wilcoxonTestForCols: (Column*) c1 : (Column*) c2 {
	if( !c1 || !c2 ) return;
	if( c1 == c2 )
		[NSException raise: NSInvalidArgumentException 
			format: @"-wilcoxonTestForCols: got equal columns"];
SX_SET_LOCALE
	wilcoxon_test([c1 getArray], [c2 getArray], [c1 count]);
SX_RESET_LOCALE
}

// Chi^2-Test auf gleiche Haeufigkeiten
-(void) equalFrequTest: sender {
	Column* c1;
	NSArray* cols = [self getCols: 1];
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	[self equalFrequTestForCol: c1];
}
-(void) equalFrequTestForCol: (Column*) col {
	if( !col ) return;
SX_SET_LOCALE
	equal_freq([col getArray], [col count]);
SX_RESET_LOCALE
}

// Chi^2-Test auf Gleicheit von gemessenen mit theor. Haeufigkeiten
// 1st column: measured distribution
// 2nd column: expected distribution
-(void) compareFrequTest: sender {
	Column* c1, *c2;
	NSArray* cols = [self getCols: 2];
	if( !cols ) return;
	c1 = (Column*)[cols objectAtIndex: 0];
	c2 = (Column*)[cols objectAtIndex: 1];
	[self compareFrequTestMeasured: c1 expected: c2];
}
-(void) compareFrequTestMeasured: (Column*) m expected: (Column*) e {
	if( !m || !e ) return;
	if( m == e )
		[NSException raise: NSInvalidArgumentException 
			format: @"-compareFrequTestForCols: got equal columns"];
SX_SET_LOCALE
	compare_freq([m getArray], [m count], [e getArray], [e count]);
SX_RESET_LOCALE
}

//============================= END STATISTCS ==============================
//==========================================================================

-(void) dealloc {
	[manager release];
	[super dealloc];
}
@end

#undef SX_SET_LOCALE
#undef SX_RESET_LOCALE
