/* This file is part of StatistX
**
** It is distributed under the GNU General Public License.
** See the file COPYING for details.
** 
** (c) Andreas Beyer, 2000
** abeyer@usf.uni-osnabrueck.de
**
**  http://www.usf.Uni-Osnabrueck.DE/~abeyer/private/StatistX/
**  
***************************************************************/

/* SXSettings.m
 * SXSettings manages general settings of the publications
 * like pathnames, etc.
 */
#include "SXSettings.h"

#include <Foundation/NSArray.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSFileManager.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>

#include <GToolKit/GToolKit.h>

#include "sx_defs.h"
#include "BrowseButton.h"
#include "Message.h"
#include "Manager.h"
#include "SXException.h"

@implementation SXSettingsEntry 
{
	NSString *	_key;
	NSString *	_value;
	NSString *	_description;
	// determines the behavior of the dialog
	int		_value_type;
	// widget allowing the user to set the parameter
	GTKWidget * _widget;
	GTKEntry  * _entry;
}
+entryWithKey: (NSString*) aKey 
	 value: (NSString*) aValue 
	 description: (NSString*) aDescription 
	 valueType: (int) value_type {  // either 'SXVTstring' or 'SXVTpath'
	return [[[self alloc] 
		initWithKey: aKey value: aValue description: aDescription 
			valueType: value_type]
		autorelease];

}
+entryFromString: (NSString*) descr {
	return [[[self alloc] initFromString: descr ] autorelease];

}
-initWithKey: (NSString*) aKey 
	 value: (NSString*) aValue 
	 description: (NSString*) aDescription 
	 valueType: (int) value_type {  // either 'SXVTstring' or 'SXVTpath'
	[super init];
	_key = [aKey retain];
	_value = [aValue retain];
	_description = [aDescription retain];
	_value_type = value_type;
	// special treatment for LOG_FILE
	if( ([_key caseInsensitiveCompare: @"LOG_FILE"] == NSOrderedSame) 
	&&  ([_value length] > 0) ) {
		[[Manager getManager] setLogFileName: _value];
	}
	return self;
}
// init with description as created by -stringRepresentation
-initFromString: (NSString*) descr {
	NSArray	* lines = [descr componentsSeparatedByString:@"\n"]; 
	NSString	* line;
	NSRange	range;
	NSString	*key, *value, *description = NULL;
	int		type;
	unsigned	n = [lines count];
	if( n == 2 ) { // handle description
		line = (NSString*) [lines objectAtIndex: 0];
		if( [line characterAtIndex: 0] == '#' ) {
			description = [[line substringFromIndex: 1] retain];
		} else {
			[NSException raise: NSInvalidArgumentException		
				format: @"-initFromString: Illegal format in string: \"%@\"", 
				descr];
		}
		line = (NSString*) [lines objectAtIndex: 1];
	} else {
		description = [NSString new];
		line = (NSString*) [lines objectAtIndex: 0];
	}
	// handle data
	range = [line rangeOfString: @"\t"];
	if( !range.length )
		[NSException raise: NSInvalidArgumentException		
			format: @"-initFromString: Illegal format in string: \"%@\"", 
			descr];
	key = [[line substringToIndex: range.location] retain]; 
	line = [line substringFromIndex: range.location+range.length];
	range = [line rangeOfString: @"\t"];
	if( !range.length ) { // no type given
		value = [line retain];	
		type = 0; // set default type
	} else { // type definition at end of line
		value = [[line substringToIndex: range.location] retain];
		type = [[line substringFromIndex: range.location+range.length]
			intValue];
	}
/*
fprintf(stderr, "desc: %s\n", [description cString]);
fprintf(stderr, "key: %s\n", [key cString]);
fprintf(stderr, "value: %s\n", [value cString]);
fprintf(stderr, "type: %d\n", type);
*/
	return [self initWithKey: key value: value 
		description: description valueType: type];
}

-(NSString*)key {
	return _key;
}
-(NSString*)value {
	return _value;
}
-(NSString*)description {
	return _description;
}
-(int) valueType {
	return _value_type;
}
/** Returns a *new* widget allowing the use to set the parameter.
    The calling object is responsible for releasing the widget. */
-(GTKWidget*)widget {
	BrowseButton * button;
	[_widget release];
	[_entry release];
	_widget = [[GTKHBox hBoxWithHomogeneous: NO spacing:0] retain];
	_entry  = [GTKEntry new];
	[_entry setText:_value];
	[(GTKBox*)_widget packStart:_entry expand:YES fill:YES padding:0];
	if( _value_type == SXVTpath ) {
		button = [BrowseButton buttonWithLabel: @"Browse" delegate: _entry];
		[(GTKBox*)_widget packStart:button expand:YES fill:YES padding:0];
	}
	return _widget;
}
/** Update the value with the current value of the widget. */
-(void)takeValueFromWidget {
	[_value release];
	_value = [[_entry stringValue] retain];
}

/** Return a string representation of the contents.
    The result can be used to initialize a new SXSettingsEntry 
    using initFromString:
  */
-(NSString*)stringRepresentation {
	return [NSString stringWithFormat: @"#%@\n%@\t%@\t%d", 
			_description, _key, _value, _value_type];
}

-(void) dealloc {
	[_key release];
	[_value release];
	[_description release];

	[super dealloc];
}
@end

static SXSettings * _settings;

@implementation SXSettings
/** The proper way to obtain the *unique* SXSettings-object: */
+(SXSettings*)settings {
	if( _settings ) return _settings;
	_settings = [self new];
	return _settings;
}
-init {
	if( _settings )
		[NSException raise: NSGenericException
			format: @"-init: illegal call of init"];
	_settings = [super init];
	dialog = NULL;
	dictionary = [NSMutableDictionary new];
	col_separators = @SX_COL_SEPARATORS;
	settings_file = [[@STD_RC_FILE stringByExpandingTildeInPath] retain];
	// allways set defaults to ensure that all values are defined
	[self setDefaults];
	if( ! [self readSettingsFromFile: settings_file] ) {
	// create a new settings file with defaults
		[self saveSettings];
	}
	[[Manager getManager] registerAction: @"/_Edit/_Preferences"
			target: self selector: @selector(raiseSettingsDialog)];
	return self;
}
-(BOOL)readSettingsFromFile: (NSString*) file_name {
	NSString * file_cont = [NSString stringWithContentsOfFile: file_name];
	NSArray  * lines = [file_cont componentsSeparatedByString:@"\n"];
	NSString * line;
	unsigned   i, cnt = [lines count];
	if( cnt < 2 ) return NO; // Can't open file
	for( i = 0; i < cnt; i+=2 ) {
		SXSettingsEntry * entry;
		NSString * arg;
		if( i > cnt-2 ) break; // invalid file format
		line = [lines objectAtIndex: i];
		arg  = [NSString stringWithFormat: @"%@\n%@", 
					line, [lines objectAtIndex: i+1]];
		entry = [SXSettingsEntry entryFromString: arg];
		[self addObject: entry];
	}
	return YES;
}
// invokes writeSettingsToFile with the default file name
-(void)saveSettings {
	NSString * msg;
	if( ![self writeSettingsToFile: settings_file] ) {
		msg = [NSString stringWithFormat: 
			@"Can't write preferences to %@", settings_file];
		[Message messageWithContents: msg title: @"StatistX: Error"];
	} else {
		msg = [NSString stringWithFormat: 
		@"Wrote preferences to %@\nIt is suggested to restart StatistX.",
			settings_file];
		[Message messageWithContents: msg title: @"Settings Written"];
	}
}
-(BOOL)writeSettingsToFile:  (NSString*) file_name {
	NSMutableString * file_cont = [NSMutableString new];
	NSArray * values = [dictionary allValues];
	unsigned i, cnt = [values count];
	for( i = 0; i < cnt; i++ ) {
		SXSettingsEntry * entry = [values objectAtIndex: i];
		[file_cont appendFormat: @"%@\n", [entry stringRepresentation]];
	}
	[[NSFileManager defaultManager] 
		removeFileAtPath: file_name handler:NULL];
	return [file_cont writeToFile: file_name atomically: NO ];
}

/** Set the default parameters */
-(void)setDefaults {
	unsigned i;
	for( i = 0; i < n_of_defaults; i++ ) {
	   [self addObject: [SXSettingsEntry 
			entryWithKey: DefaultSettings[i].key
				 value: DefaultSettings[i].value
			 description: DefaultSettings[i].description
			   valueType: DefaultSettings[i].value_type
			]
	   ];
	}
}

-(void)takeSettingsFromDialog {
	NSArray * values = [dictionary allValues];
	unsigned i, cnt = [values count];
	for( i = 0; i< cnt; i++ ) {
		[(SXSettingsEntry*)[values objectAtIndex: i]
			takeValueFromWidget];
	}
}

/** Access hashtable: */
/** Return number of entries in dictionary */
-(unsigned)count {
	return [dictionary count];
}

/** Returns the whole SXSettingsEntry object */
-(SXSettingsEntry*)objectForKey: aKey {
	return [dictionary objectForKey: aKey];
}
/** Returns the value containes in the according SXSettingsEntry 
    object as a string */
-(NSString*)valueForKey: aKey {
	return [[dictionary objectForKey: aKey] value];
}
/** Add SXSettingsEntry object to dictionary. */
-(void)addObject: (SXSettingsEntry*)anObject {
	[dictionary setObject: anObject forKey:[anObject key]]; 
}
-(void)removeObjectForKey: aKey {
	[dictionary removeObjectForKey: aKey];
}

/** Return the default separators for data files */
-(NSString*)columnSeparators {
	return col_separators;
}
/** return directory containing data files */
-(NSString*) dataDir {
	return [self valueForKey: @"DATA_DIR"];
}

-(void) cancel {
	[GTKApp stopModal];
	[dialog destroy];
}

-(void) ok {
	[GTKApp stopModal];
	[dialog destroy];
	[self takeSettingsFromDialog];
	[self saveSettings];
}

// opens a sttings dialog
-(void)raiseSettingsDialog {
	NSArray	*values = [dictionary allValues]; 
	GTKVBox	*v_box;
	GTKHBox	*buttons;
	GTKTable	*table;

	GTKButton	*ok, *cancel;
	SXSettingsEntry * entry;
	GTKWidget	*widget;
	GTKLabel	*label;
	unsigned	i, cnt = [values count];
	
	[dialog release];
	dialog = [GTKDialog new];
	
	ok = [GTKButton buttonWithLabel: @"Save Settings"];
	[ok setUsize: 100 height: 30];
	[ok connectSignal:@"clicked" withAction:
        	[GTKAction actionWithTarget:self selector:@selector(ok)]];
	cancel = [GTKButton buttonWithLabel: @"Cancel"];
	[cancel connectSignal:@"clicked" withAction:
        	[GTKAction actionWithTarget:self selector:@selector(cancel)]];
	[cancel setUsize: 100 height: 30];

	buttons = (GTKHBox*) [dialog actionArea];
	[buttons packStart:ok expand:NO fill:NO padding:0];
	[buttons packStart:cancel expand:NO fill:NO padding:0];

	v_box = (GTKVBox *) [dialog vbox];
	[v_box setBorderWidth: 5];
	table = [GTKTable tableWithRows: cnt columns: 2 homogeneous: NO ];
	for( i = 0; i < cnt; i++ ) {
		entry = (SXSettingsEntry*) [values objectAtIndex: i];
		label = [GTKLabel labelWithStr: [entry description]];
		[label setAlignment: 1 yalign: 0];
		widget = [entry widget];
		[table attach: label
			leftAttach:  0
			rightAttach: 1 
			topAttach:   i
			bottomAttach:i+1
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 5
			ypadding: 1
		];
		[table attach: widget
			leftAttach:  1
			rightAttach: 2 
			topAttach:   i
			bottomAttach:i+1
			xoptions: GTK_EXPAND | GTK_FILL
			yoptions: GTK_EXPAND
			xpadding: 1
			ypadding: 1
		];

	}	
	[v_box packStart:table  expand:YES fill:YES padding:0];

	[dialog setTitle: @"StatistX Preferences"];
	[dialog setPosition: GTK_WIN_POS_CENTER ];
	[dialog show];
	[GTKApp runModalForWindow: dialog];
}

-(void) dealloc {
	[dialog release];
	[dictionary release];
	[settings_file release];	
	[super dealloc];
}
@end
